#!/usr/bin/perl
#
# Top-level perl script for CSPP_HEAP package.
#
# Uses only Perl modules that are part of the v5.010 standard install.
#
# A set of user-defined functions are appended to this main script;
# they are listed in alphabetical order.
#
# Pre-requisites:
# HEAP distribution in directory $CSPP_HEAP_HOME/$CSPP_HEAP_REVNAME
#
# START-HELP (N.B. This tag is used in printing help, do not touch!)
#
#  Options:
#  --version             Show program's version number and exit
#  ?, -h, --help         Show this help message and exit
#  -n, --nothing         Make the PCF files, but do not execute.
#
#  Mandatory Arguments:
#    At a minimum these arguments must be specified [i.e. no default].
#
#    -s MISSION, --sat=MISSION
#                        The mission satellite data for HEAP algorithm.
#                        Possible values are: npp, j01, m01, m02, m03.
#    -i INPUT_FILES, --input_files=INPUT_FILES
#                        Path to input files. May be directory name, or
#                        directory/filemask (use * and ?, not regexp).
#                        Wildcarded filemasks MUST be "quoted" (or OS expands).
#                        e.g. -i "/dataDir/{GATMO,GCRSO,SCRIS,TATMS}*t18*.h5",
#                        or -i "/dataDir/{IASI_xxx,amsual1b,mhsl1b}*20200504*".
#                        (BUT -i /dataDir  really is the simplest way).
#
#  Extra Options:
#    These options may be used to customize behaviour of this program.
#
#    -m MODE, --mode=MODE
#                        Mode of processing of form X_Y with different options
#                        for JPSS and MetOp MISSIONs. Y is IRMW for infrared +
#                        microwave (JPSS: CrIS+ATMS; MetOp: IASI+AMSUA/MHS) or
#                        IRO for infrared-only. For JPSS, X is LR or HR, where
#                        LR means NSR (SCRIS) and HR means FSR (SCRIF). X for
#                        MetOp selects how to disaggregate IASI L1C files into
#                        22/23 MDRs per granule; S to include overpass start
#                        line, E to include end line, and M to prefer the middle
#                        of overpass. S,M,E matter only for DB overpasses; they 
#                        are ignored, or can be omitted, for granular IASI L1C.
#                        [default: -mode=HR_IRMW (JPSS), -mode=M_IRMW (MetOp)].
#    -o OUTPUT, --output=OUTPUT
#                        Output option string of form OSR where O is for OLR
#                        (outgoing longwave radiation), S is for Subsetter, &
#                        R is for Retrieval. Any combination of these letter
#                        options is valid.
#                        [default: --output==R].
#    -w WORK_DIR, --work=WORK_DIR
#                        The directory in which all activity will occur, this
#                        is also where output files will be written.
#                        [default: current directory].
#    -d DYNAMIC_ANC, --dynanc=DYNAMIC_ANC
#                        Path to root of dynamic ancillary data tree; beneath
#                        here data are organized as per SSEC remote ancillary
#                        server for CSPP, i.e. %Y_%m_%d_%j where %Y is 4-digit
#                        year, %m is 2-digit month, %d is 2-digit day of month
#                        and %j is 3-digit day of year.
#                        [default: environment var. $CSPP_DYNAMIC_ANCIL_DIR].
#    -f N_FOVS, --fovs=N_FOVS
#                        The number of fields of view per CrIS field of regard.
#                        Added to accommodate loss of FOV.4 from DB data stream
#                        starting 1-Dec-2014. For IDPS data, and DB data before
#                        01-Dec-2014, you can use --fovs=9 (or -f 9) and for DB
#                        data before 18-Aug-2015 use --fovs=8 (or -f 8).  This
#                        has no meaning for MetOp missions.
#                        [default: --fovs=7 for npp, --fovs=9 for j01].
#    -p PROCESSORS, --processors=PROCESSORS
#                        Number of cpus to use, parallel only within overpass.
#                        [default: --processors=1].
#    -v, --verbose       Each occurrence increases verbosity 1 level from
#                        [default: ERROR only] -v=WARNING -vv=INFO -vvv=DEBUG
#    -c, --clean         Deletes temporary working directory under WORK_DIR
#                        at process completion.
#                        [default: --clean; --noclean is the contrary option].
#    -q, --quarantine    Processing machine is not connected to the web for the
#                        purpose of fetching required dynamic ancillary data.
#                        [default: --noquarantine].
#    -a, --aggregate     Aggregates NUCAPS-EDR and NUCAPS-CCR-AR output files
#                        into swaths of up to 30 minutes as a POST-PROCESSING
#                        step.  Variables are reshaped to a non-standard form
#                        and files named NUCAPS-AGG-EDR & NUCAPS-AGG-CCR-AR to
#                        distinguish from standard format & length granules.
#                        [default: --agg, --noagg is the contrary].
#    -b, --break         Breaks (disaggregates) 'npp' and 'j01' SDRs into 32s
#                        granules as a PRE-PROCESSING step. If input SDRs are
#                        provided ALREADY as 32-second granules, choosing the
#                        contrary option, --nobreak, saves both time and
#                        temporary disk space.  No effect on MetOp processing.
#                        [default: --break, --nobreak or -nob is the contrary].
#    -x, --xfiles        Use non-DB filenames for AMSUA/MHS L1B data sources.
#                        DB filenames are like {amsua,mhs}l1b_M0{1,2,3}_*.l1b.
#                        Non-DB filenames are like *.{AMAX,MHSX}.M{1,2,3}.* and
#                        under this option IASI L1C files are assumed granular,
#                        i.e. comprise 22/23 MDRs per file.
#                        [default: -nox, i.e. Direct Broadcast filenames].
#
# END-HELP (N.B. This tag is used in printing help, do not touch!)
#
# The following support is now hidden - the default is 3, don't be using
# anything else unless you are sure you know what you are doing.
#
#    -g GFS, --gfs=GFS   Reprocessing mode using GFS data in grib version 1 or 2,
#                        where GFS is 1 or 2. Activating this option also sets
#                        the GFS local repo to DYNAMIC_ANC/%Y/%m (see above)
#                        AND forces --quarantine (see below).  ALSO NOTE:
#                        --gfs=0 means use 1.0-degree GFS in grib2 format,
#                        --gfs=3 means use 0.5-degree GFS in grib2 format
#                        In both these cases (0 & 3) reprocessing mode is OFF.
#                        [default: --gfs=3].
#    
# NOTES-TO-SELF:
# -------------
#   $!=N to change exit code on die() to N; default is 255
#
#   --clean removes the temporary directory tree. If the user selects --noclean
#   the temporary directory is NOT removed and all HEAP product files are 
#   available in the temporary directory structure. This option might be useful
#   for debugging and is necessary if the user wants access to the product files
#   generated prior to the output NetCDF files (that are copied up to WORK_DIR).
#   It is then user responsibility to dispense with unwanted directories/files.
#
# This is a valid way to run a single S-NPP granule ...
#
#   ./run_heap.pl -s npp -m LR_IRMW \
#     -i "/dataDir/{GATMO,GCRSO,SCRIS,TATMS}_npp_d20131120_t0656260_*_cspp_dev.h5"
#
# The quotes are to prevent OS expanding wildcards/tokens on the commandline.
#
# A design decision is to require input data to be in a single directory.
# See use of $options{input_dir} in function validate_options().
#
#
# DESIGN PHILOSOPHY
# -----------------
# Build up what you need to run the process in a temporary working directory,
# using symlinks where you can to minimise i/o and duplication. Because you
# could have multiple swaths of data in INPUTFILES, create a swath_nnn below
# temporary working directory for each swath. The typical DB use case is going
# to be a single swath per processing run (I think). I am going to build the
# case for, and process, swath_001 before I even start building the case for
# swath_002 (if it exists). This, and the fact that processing is done granule
# by granule within a swath, has implications for abort conditions - I do not
# want to abort the top-level script if there is any chance that subsequent
# granules and swaths COULD complete successfully.
#
# The over-arching idea is that, as you develop, wherever you get to, or 
# wherever it fails, there can be kept a temporary directory where you can go
# and run interactively with all required inputs and control files present.
# This helps development, debugging, and in investigating user error post-
# distribution.
#
# A note on aggregation:  You cannot aggregate pre-processing and have HEAP
# work; you may even need to dis-aggregate (with nagg for JPSS), in fact that
# is what I now do (i.e. disaggregate) whether it is needed or not; the typical
# use case has CrIS/ATMS SDR files as overpass-length aggregates. This
# means I must use the CSPP patched nagg which knows how to disaggregate SCRIF
# files!  For MetOp IASI L1C I have to disggregate with kai otherise it just
# won't work.  I had to patch iasi_level1c_reader.cpp to change a FATAL to
# a NOTICE in two places. The DB character of our DB data is not divined in
# iasi_level1c_reader.cpp (even though there is clearly an intent to do so by
# examining a bit or flag in the IASI L1C file), so a workaround is needed.
# It is exactly the same patch that I made to the same source in NUCAPS-IASI.
#
# The final piece of the puzzle is to post-aggregate the NUCAPS-CCR-AR_* and
# NUCAPS-EDR_* 4-scan granules into overpass length files before posting
# up to WORK_DIR. This has been a bit troublesome as the 4-scan granules have
# been flattened to one horizontal dimension; but it makes little sense to just
# grow this one dimension when aggregating the retrievals into swaths. So I
# have reshaped as 30 FORs by as many scanlines as are present (4 per NUCAPS
# output netCDF4 file before post-aggregation).  So dimensions had to change.
# Also, in order to display the 3D (profile) variables in McIDAS-V, I needed to
# have a single vertical-levels profile, so 3D Pressure and Effective_Pressure
# were recast as 1D and all the duplicates were discarded (at every 2D data
# point, we did have the same set of pressure levels).  I have not changed the
# following 3D fields to be McIDAS-V imaging compatible: Cloud_Top_Pressure,
# Cloud_Top_Fraction.  There are 8 cloud levels supported and I would have to
# split both of these into 8 separately named 2D fields (Cloud_Top_Pressure_1,
# Cloud_Top_Pressure_2, etc), and that is messy. And because of how these data
# are stored, I would also have to do much more coding to achieve (all the
# other fields are processed analagous to running through sed and changing a
# couple of things; Cloud_Top_Pressure & Cloud_Top_Fraction need reshaping
# and managing in memory before writing out - a very different pattern).
# NOTE: The above might be a dangerous strategy if pressure levels change
# to be specific to the pixel. Better to save presures as a 3D cube and then
# figure out how to display later - fidelity more important than convenience.
#
# Aronne suggested a sensible rule for aggregation.  If the user does not
# choose aggregate, the output granules are the same size as the input
# granules, otherwise they are aggregated to swath (i.e. overpass) length,
# or to 30 minutes, whichever is shorter.  At first blush, this makes a lot
# of sense. But if the input granules are aggregated ones, they have to be
# disaggregated in pre-processing (this is a NUCAPS requirement), so to get
# back to their input size, the outputs still need to be aggregated - and
# that means they have the data format of an aggregated file!  Currently,
# input files are disaggregated into one big bucket and their data times
# determine how to order into overpasses and, thus, into output files - so
# there are problems to solve here still to achieve fileOutSize==fileInSize.
# My view is that if you don't like how I aggregate in post-processing:,
# don't choose it, consult the NOAA documentation for file formats, and
# sort it out for yourself.
#
# Am I going to add post-processing aggregation for MetOp?  Maybe.
#
#
# JED  Nov-2014, updated  Jul-2015, and again in Oct-2015.
#
# JED  Oct 2017 for version 2.0, now version-controlled by git.
#
# JED  Aug-2018 for version 2.1.
#
# JED  May/June-2020 for HEAP version 1.0
#


# Pragmata
# ````````
use 5.010;
use strict;
use warnings;


# Load Perl modules - standard modules only
# `````````````````````````````````````````
use POSIX;
use Storable;
use File::Spec;
use File::Copy;
use Data::Dumper;
use Getopt::Long;
use File::Basename;
use Cwd qw(abs_path);
use Encode qw(decode);
use Time::gmtime qw(:FIELDS);
use List::Util qw(sum min max);
use File::Path qw(mkpath rmtree);
use Time::Local qw(timegm_nocheck);
use File::Temp qw(tempfile tempdir);
use Scalar::Util qw(looks_like_number);
use I18N::Langinfo qw(langinfo CODESET);


# Global variables are all UPPERCASE throughout
# `````````````````````````````````````````````
our $AGG_WINDOW_MINUTES = 30; # 30-minute aggregation window for npp
our $SLEEP_GFS_MINUTES = 0;   # Disable wait for preferred GFS data
our $MAIN = basename($0);
defined $ENV{CSPP_HEAP_HOME} or die "$MAIN\:: CSPP_HEAP_HOME not set.\nAbort\n";
defined $ENV{CSPP_HEAP_REVNAME} or die "$MAIN\:: CSPP_HEAP_REVNAME not set.\nAbort\n";
defined $ENV{CSPP_HEAP_REVDATE} or die "$MAIN\:: CSPP_HEAP_REVDATE not set.\nAbort\n";
defined $ENV{JPSS_REMOTE_ANC_DIR} or die "$MAIN\:: JPSS_REMOTE_ANC_DIR not set.\nAbort\n";
defined $ENV{CSPP_DYNAMIC_ANCIL_DIR} or die "$MAIN\:: CSPP_DYNAMIC_ANCIL_DIR not set.\nAbort\n";
our $OPS = File::Spec -> catfile($ENV{CSPP_HEAP_HOME}, $ENV{CSPP_HEAP_REVNAME}, 'OPS');
(-d $OPS) or die "$MAIN\:: CSPP_HEAP subdirectory $OPS not found.\nAbort";
our $JPSS_REMOTE_ANC_DIR = $ENV{JPSS_REMOTE_ANC_DIR};
our $CSPP_DYNAMIC_ANCIL_DIR = $ENV{CSPP_DYNAMIC_ANCIL_DIR};


# $H5DUMP points to supplied static-compiled h5dump (version 1.8.11)
# ``````````````````````````````````````````````````````````````````
our $H5DUMP = File::Spec -> catfile($ENV{'CSPP_HEAP_HOME'}, 'bin', 'h5dump');
(-e $H5DUMP) or die "$MAIN\:: h5dump $H5DUMP not found.\nAbort";
(-x $H5DUMP) or die "$MAIN\:: h5dump $H5DUMP not executable.\nAbort";


# $H5REPACK points to supplied static-compiled h5dump (version 1.8.11)
# ````````````````````````````````````````````````````````````````````
our $H5REPACK = File::Spec -> catfile($ENV{'CSPP_HEAP_HOME'}, 'bin', 'h5repack');
(-e $H5REPACK) or die "$MAIN\:: h5repack $H5REPACK not found.\nAbort";
(-x $H5REPACK) or die "$MAIN\:: h5repack $H5REPACK not executable.\nAbort";


# $WGRIB points to supplied static-compiled wgrib (v1.8.1.2a)
# ```````````````````````````````````````````````````````````
our $WGRIB = File::Spec -> catfile($ENV{'CSPP_HEAP_HOME'}, 'bin', 'wgrib');
(-e $WGRIB) or die "$MAIN\:: h5repack $WGRIB not found.\nAbort";
(-x $WGRIB) or die "$MAIN\:: h5repack $WGRIB not executable.\nAbort";


# $WGRIB2 points to supplied static-compiled wgrib2 (v0.1.9.5.1)
# ``````````````````````````````````````````````````````````````
our $WGRIB2 = File::Spec -> catfile($ENV{'CSPP_HEAP_HOME'}, 'bin', 'wgrib2');
(-e $WGRIB2) or die "$MAIN\:: h5repack $WGRIB2 not found.\nAbort";
(-x $WGRIB2) or die "$MAIN\:: h5repack $WGRIB2 not executable.\nAbort";


# $H5NAGG points to supplied static-compiled nagg
# ```````````````````````````````````````````````
# But you can point to your own nagg if you wish. Ours is Version 1.5.1
our $H5NAGG = File::Spec -> catfile($ENV{'CSPP_HEAP_HOME'}, 'bin', 'nagg');
(-e $H5NAGG) or die "$MAIN\:: nagg $H5NAGG not found.\nAbort";
(-x $H5NAGG) or die "$MAIN\:: nagg $H5NAGG not executable.\nAbort";


# $KAI points to supplied dymamic-compiled kai
# ````````````````````````````````````````````
# Compiled from kai-1.11.tar.gz supplied by Simon.Elliott@eumetsat.int
our $KAI = File::Spec -> catfile($ENV{'CSPP_HEAP_HOME'}, 'bin', 'kai');
(-e $KAI) or die "$MAIN\:: kai $KAI not found.\nAbort";
(-x $KAI) or die "$MAIN\:: kai $KAI not executable.\nAbort";


# Get the command line options, display help or revision if asked
# ```````````````````````````````````````````````````````````````
my $revision = "CSPP_HEAP v1.0 $ENV{CSPP_HEAP_REVDATE}";
my $pwd = getcwd();
my %options = get_commandline($revision, $pwd);
($options{verbose} >= 2) and pretty_dump('%options', %options);


# uuidgen is a system requirement
# ```````````````````````````````
our $UUIDGEN = `which uuidgen`; chomp($UUIDGEN);
(index($UUIDGEN, 'uuidgen') < 0) and
  die "$MAIN\:: Required utility uuidgen not found\nAbort"; 


# At present we are saying that lftp is a system requirement
# ``````````````````````````````````````````````````````````
our $LFTP = 'lftp';
($options{quarantine}) or do {
  system("$LFTP --version 1>/dev/null 2>/dev/null") == 0
    or die "$MAIN\:: Required utility lftp not found\nAbort"; };


# $NCDUMP points to supplied static-compiled ncdump (netcdf library version 4.3.0)
# ````````````````````````````````````````````````````````````````````````````````
our $NCDUMP = File::Spec -> catfile($ENV{'CSPP_HEAP_HOME'}, 'bin', 'ncdump');
($options{aggregate}) and do {
  (-e $NCDUMP) or die "$MAIN\:: $NCDUMP not found.\nAbort";
  (-x $NCDUMP) or die "$MAIN\:: $NCDUMP not executable.\nAbort"; };


# $NCGEN points to supplied static-compiled ncgen (netcdf library version 4.3.0)
# ``````````````````````````````````````````````````````````````````````````````
our $NCGEN = File::Spec -> catfile($ENV{'CSPP_HEAP_HOME'}, 'bin', 'ncgen');
($options{aggregate}) and do {
  (-e $NCGEN) or die "$MAIN\:: $NCGEN not found.\nAbort";
  (-x $NCGEN) or die "$MAIN\:: $NCGEN not executable.\nAbort"; };


# $HEAPAGG points to perl script to post aggregate EDR and CCR-AR files
# `````````````````````````````````````````````````````````````````````
our $HEAPAGG = File::Spec -> catfile($ENV{'CSPP_HEAP_HOME'}, 'scripts',
  'aggregate_swath.pl');
($options{aggregate}) and do {
  (-e $HEAPAGG) or die "$MAIN\:: $HEAPAGG not found.\nAbort";
  (-x $HEAPAGG) or die "$MAIN\:: $HEAPAGG not executable.\nAbort"; };


# Enumerate the MetOp filenaming options
# ``````````````````````````````````````
my $metopA = '{{amsua,mhs}l1b_M02_*.l1b,IASI_xxx_1C_M02_*}';
my $metopB = '{{amsua,mhs}l1b_M01_*.l1b,IASI_xxx_1C_M01_*}';
my $metopC = '{{amsua,mhs}l1b_M03_*.l1b,IASI_xxx_1C_M03_*}';
($options{xfiles}) and do {
  $metopA = '{*.{AMAX,MHSX}.M2.*,IASI_xxx_1C_M02_*}';
  $metopB = '{*.{AMAX,MHSX}.M1.*,IASI_xxx_1C_M01_*}';
  $metopC = '{*.{AMAX,MHSX}.M3.*,IASI_xxx_1C_M03_*}'; };


# These are used for fileglobbing when INPUTFILES is a directory name
# ```````````````````````````````````````````````````````````````````
my %filemask = (j01 => '{GATMO,GCRSO,SCRIS,TATMS}_j01_*.h5',
                npp => '{GATMO,GCRSO,SCRIS,TATMS}_npp_*.h5',
                m02 => $metopA,
                m01 => $metopB,
                m03 => $metopC);
# Distinguish SCRIS from SCRIF because they might be in same directory!
(index($options{mode}, 'HR') < 0) or do {
  $filemask{j01} =~ s/SCRIS/SCRIF/g;
  $filemask{npp} =~ s/SCRIS/SCRIF/g; };
# Add in cloud mask & height files.
if (index($options{output}, 'C') != -1) {
  $filemask{j01} =~ s/TATMS/TATMS,GMTCO,JRR-Cloud*/g;
  $filemask{npp} =~ s/TATMS/TATMS,GMTCO,JRR-Cloud*/g;
  $filemask{j01} =~ s/\.h5/\.{h5,nc}/g; # because the JRR are .nc
  $filemask{npp} =~ s/\.h5/\.{h5,nc}/g;
}
(index($options{mode}, 'IRO') < 0) or do {
  $filemask{j01} =~ s/GATMO,//g;
  $filemask{j01} =~ s/,TATMS//g;
  $filemask{npp} =~ s/GATMO,//g;
  $filemask{npp} =~ s/,TATMS//g; };
# Debug information
($options{verbose} == 3) and pretty_dump('%filemask', %filemask);


# Parse and validate command line %options & make temporary working directory
# ```````````````````````````````````````````````````````````````````````````
my $sat = $options{sat};
%options = validate_options($filemask{$sat}, %options);
($options{verbose} >= 3) and pretty_dump('%options', %options);
my $tmpdir = tempdir('TMP_XXXXXX', DIR => $options{work},
  CLEANUP => $options{clean});
chdir($tmpdir) or die "$MAIN\:: Cannot chdir to temp dir $tmpdir\nAbort";
($options{verbose} >= 3) and print "chdir to working directory $tmpdir\n";


# Make the template PCF files at the $tmpdir level (this is $PWD)
# ```````````````````````````````````````````````````````````````
# Policy: I want to make the templates with as much information in them
# as I know right now - so stuff like the satellite, whether clouds are
# included, microwave included or not, etc.  But for things I do not know
# now - like the specific granule filenames, the job start and end times,
# the granuleinfo txt filename - instead they have place holders ready for
# replacing with granule-specific information later.
#
foreach my $step (split //, $options{output}) {
  make_heap_pcf_template($step, %options);
}


# Preprocess and stage data for HEAP system compatibility
# ```````````````````````````````````````````````````````
# I have added logic to disaggregate GMTCO files if Clouds selected, and to
# copy the JRR-CloudMask/Height files across into input/. It gets ugly - the
# ATMS files have one set of file timecodes, the CRIS files have another, and
# now the VIIRS/JRR have yet another.  There is logic to deal with lining up
# matching CRIS, ATMS and VIIRS files [in function populate_swath_info_jpss()],
# with results put in the %granule hashes that make up @swath. For ATMS
# and CRIS, I have to figure which files I need, but it is always 3 ATMS
# and 1 CRIS file per PCF.  For VIIRS/JRR I need to check tiems for oevrlap
# and it could be one or more files that need to be slotted into PCFs.
# MetOp are also dealt with in cspp_preprocess() but a separate function is
# provided [populate_swath_info_metop()] to load %granule hashes for each
# @swath.
%options = cspp_preprocess(%options);
($options{verbose} >= 1) and pretty_dump('%options', %options);


# Define dynamic data remote & local locations and file naming conventions
# ````````````````````````````````````````````````````````````````````````
# The root of the local directory for dynamic ancillary data is either defined
# on the command line or is assigned a default value. I am appending the agreed
# subdirectory naming convention to this root and copying it across to a hash
# that contains much of what I need to know to fetch and store ancillary data.
# See https://jpssdb.ssec.wisc.edu/cspp_v_2_0/ancillary/ for directory structure.
# I have added support for GRIB1 files and an alternative repo directory for
# reprocessing.  Also added support for 1/2 GFS 
my $usual_repo = File::Spec -> catfile($options{dynanc}, '%Y_%m_%d_%j');
my $repro_repo = File::Spec -> catfile($options{dynanc}, '%Y/%m');
my $local_repo = $usual_repo;
($options{gfs} == 1) and $local_repo = $repro_repo;
($options{gfs} == 2) and $local_repo = $repro_repo;
my $grib1_template = 'gblav.PGrbF%f.%y%m%d.%Hz';
my $grib2_template = 'gfs.t%H.%y%m%d.pgrbf%f';
($options{gfs} == 3) and $grib2_template = 'gfs.t%Hz.pgrb2.0p50.f0%f.%Y%m%d';
my $gfs_template = $grib2_template;
($options{gfs} == 1) and $gfs_template = $grib1_template;


# Make a hash to hold the information as I like to structure it
# `````````````````````````````````````````````````````````````
my @parts = split(/\:\/\//, $JPSS_REMOTE_ANC_DIR);
my $gfsproto = $parts[0]."://";
@parts = split(/\//, $parts[1]);
my $gfssite = shift(@parts);
my $gfsremote = join('/', @parts, basename($usual_repo));
my %dynanc_loc = (GFSPROTO => $gfsproto,
                  GFSSITE => $gfssite,
                  GFSREMOTE => $gfsremote,
                  GFSTEMPLATE => $gfs_template,
                  GFSLOCAL => $local_repo);
($options{verbose} == 3) and pretty_dump('%dynanc_loc', %dynanc_loc);


# Main processing loop in which $PWD is cycled through swath_nnn
# ``````````````````````````````````````````````````````````````
# Actually swath_ijk may exist and be loaded with links to satellite data,
# but be skipped in processing because populate_swath_info_jpss/metop() does
# not find a grouping of valid data to process.  If the --noclean has been
# set on command line, the temporary working directory will remain when this
# script terminates. In there will be bash scripts HEAP_Preprocessor.bash
# and HEAP_Retrieve.bash that you could use for debugging, say. They will
# just run the last made PCFs, i.e. the last granule.
# To add parallel processing of granules, I need to have a separate directory
# not just for each swath, but for each granule.  I want to add as little code
# as possible to do this, and to copy the minimum necessary into the granule
# directory.
foreach my $wrkdir (@{$options{process_dir}}) { # $wrkdir is the swath
  chdir("$tmpdir/$wrkdir") or die "$MAIN\:: Cannot chdir to $wrkdir\nAbort";
  store(\%options, 'options.store') or die "$MAIN\:: Can't write options.store\n";
  #
  # Define some swath variable to manage scope
  # ``````````````````````````````````````````
  my @swath = (); my $secs_per_granule = 0;
  #
  # npp and j01 CrIS + ATMS + VIIRS/JRR processing now supported
  # ````````````````````````````````````````````````````````````
  (grep {$_ eq $sat} qw/npp j01/) and do {
    @swath = populate_swath_info_jpss(\%dynanc_loc, %options);
    $secs_per_granule = 600; };
  #
  # m01, m02 and m03 with IASI + AMSUA/MHS processing now supported
  # ```````````````````````````````````````````````````````````````
  (grep {$_ eq $sat} qw/m01 m02 m03/) and do {
    @swath = populate_swath_info_metop(\%dynanc_loc, %options);
    $secs_per_granule = 3300; };
  #
  # Anything to process here?
  # `````````````````````````
  ($#swath < 0) and do { # @swath holds the granules in this swath
    ($options{verbose} >= 1) and 
      warn "WARNING: No eligible granules in $wrkdir\n"; next; };
  #
  # Loop and timeout control variables
  # ``````````````````````````````````
  my $start = time(); my @fh = (); my ($i, $j, $k) = (-1, 0, 0);
  my $timeout_secs = int(($#swath+1)/$options{processors} + 0.999) * $secs_per_granule;
  #
  # Process each granule in swath
  # `````````````````````````````
  while ($k <= $#swath) {
    $k = count_done_and_close_filehandles(@fh);
    my $n = $j - $k;
    (($n < $options{processors}) and ($i < $#swath)) and do {
      $i = $j;  $j++;  $n = $j - $k;
      my $instance = sprintf("granule_%03d", $j);
      mkpath($instance) or
        die "$MAIN\:: INSTANCE_DIR $wrkdir/$instance cannot be made.\nAbort";
      chdir($instance) or
        die "$MAIN\:: Cannot chdir to $wrkdir/$instance\nAbort";
      my %granule = %{$swath[$i]};
      store(\%granule, 'granule.store') or die "$MAIN\:: Can't write granule.store\n";
      open ($fh[$j], '|heap_granule.pl');
      chdir('..');  # chdir up from granule level to swath level
    };
    my $diff = time() - $start;
    ($diff > $timeout_secs) and do {
      warn "WARNING: Some granules may not have completed before timeout.\n";
      last; };
    sleep 1;  # I don't need this loop racing
  }
  my @movefiles = ();
  push(@movefiles, sort(glob("*.log")));
  push(@movefiles, sort(glob("*.PCF")));
  foreach (@movefiles) {
    move($_, '..') or do {
      ($options{verbose} >= 1) and warn "WARNING: Move failed: $!\n"; };
  }
  #
  # Swath complete for any MISSION, copy files up to $tmpdir
  # ````````````````````````````````````````````````````````
  # Why not move the netCDF files sooner?  Because I might be going to
  # post-aggregate (i.e. concatenate) the product files into swaths.
  # I now copy files out instead of move files out so that I have
  # the files I need in place for debugging purposes.
  ($options{nothing}) or do {
    my @copyfiles = ();
    push(@copyfiles, sort(glob("NUCAPS*.nc")));
    push(@copyfiles, sort(glob("NUCAPS*.txt")));
    push(@copyfiles, sort(glob("*.out"))); # Do I really want these?
    if ($options{aggregate} and (index($options{output}, 'R') != -1)) {
      my $cmd = "$HEAPAGG -s $sat -p EDR -i . -o ..";
      ($options{verbose} >= 1) and print "Running ... $cmd\n";
      system($cmd) == 0 or die "$MAIN\:: Could not run command $cmd\nAbort";
      $cmd = "$HEAPAGG -s $sat -p CCR-AR -i . -o ..";
      ($options{verbose} >= 1) and print "Running ... $cmd\n";
      system($cmd) == 0 or die "$MAIN\:: Could not run command $cmd\nAbort";
    }
    foreach (@copyfiles) {
      copy($_, '..') or do {
        ($options{verbose} >= 1) and warn "WARNING: Copy failed: $!\n"; };
    }
  };
  chdir($tmpdir) or die "$MAIN\:: Cannot chdir to temp dir $tmpdir\nAbort";
  #
  # Swath complete, move NUCAPS-*.nc files up to $options{work}, concat log files
  # `````````````````````````````````````````````````````````````````````````````
  # This part used to happen after all swaths complete, now I push data up to
  # $options{work} as each swath completes. If aggregation is on, there will be
  # an extra couple of files with -AGG in filename as well as the many small
  # granule-sized files.
  ($options{nothing}) or do {
    $options{clean} and system("rm -rf $wrkdir"); # save disk space on big jobs
    foreach (sort(glob("{NUCAPS*.nc,MHS-ALL*.nc,AMSUA-ALL*.nc,*.txt}"))) {
      move($_, '..') or do {
        ($options{verbose} >= 1) and warn "WARNING: Move failed: $!\n"; };
    }
  };
}


# All done, chdir to starting directory so that $tmpdir can be removed
# ````````````````````````````````````````````````````````````````````
if (not $options{clean}) {
  print "--noclean set, reminder: tmpdir is ...\n$tmpdir\n";
}
chdir($pwd) or die "$MAIN\:: Cannot chdir back to $pwd\nAbort";
exit(0);


###############################################################################
#                                                                             #
#         User-defined functions - they are listed alphabetically             #
#                                                                             #
#   NOTE: My editor can collapse code between matching {} braces, this        #
#   explains the strange formatting - I can collapse each function to see     #
#   just the function name and an example of how to call it; and expand       #
#   to see the more detailed documentation and to edit.      JED Nov 2013     #
#                                                                             #
###############################################################################

# --------------------
  sub compute_kai_skip
# ====================
# $kai_skip = compute_kai_skip(mdrstart, $iasidbfile);
  {
    # $kai_skip is the number of MDR records to skip to achieve the
    # selected $mdrstart for file $iasidbfile.
    my ($mdrstart, $file) = (shift, shift);
    my $name = (caller(0))[3];
    $name =~ s/main/$MAIN/g;
    my $skip = 0; # This sticks for ($mdrstart eq 'S')
    if (is_integer($mdrstart)) {
      $skip = $mdrstart;
    } else {
      (grep {$_ eq $mdrstart} qw/E M/) and do {
        my $cmd = "$KAI --out TMP.iasi --in $file 2>/dev/null";
        my $dribble = `$cmd`;
        unlink('TMP.iasi') or die "$name\: cannot delete TMP.iasi\nAbort";
        foreach (split("\n",$dribble)) {
          (index($_, '    MDR   :') < 0) or do {
            my $mdr = trim_string((split(':', $_))[1]);
            (is_integer($mdr)) and do {
              my $n = 0;
              my @seq = (23,22) x 99;
              while (($mdr - $seq[$n])> 0) {
                $mdr -= $seq[$n++];
              }
              ($mdrstart eq 'E') and $skip = $mdr;
              ($mdrstart eq 'M') and $skip = int($mdr/2);
            };
          };
        }
      };
    }
    return $skip;
  }

# ------------------------------------
  sub count_done_and_close_filehandles
# ====================================
# $k = count_done_and_close_filehandles(@fh);
  {
    # This is called from within the loop over granules within a swath
    # (overpass) in MAIN. @fh is the array of filehandles, the index of
    # which refers to the granules in the loop (this is 1-based indexing).
    # I do not have to close a filehandle once its granule has completed,
    # but this accumulates defunct processes that clear only when the main
    # script completes. The danger is that a process does not complete, then
    # the script will hang here waiting.
    my $k = 0;
    foreach (glob('granule_???/done.flag')) {
      $k++; my $j = int(substr($_, 8, 3));
      fileno($_[$j]) and close($_[$j]);
    }
    return($k);
  }

# -------------------
  sub cspp_preprocess
# ===================
# %options = cspp_preprocess(%options);
  {
    # On input, %options comprises some command line parameters plus other
    # processing options, including defaults for unspecified command line
    # values and extra stuff defined in main to guide processing. On output,
    # %options includes a field $options{process_dir} that points to swath
    # directories.  In this function I make swath directories and populate
    # with input data, or links to the input data, as appropriate.  I do not
    # simply link to the user supplied input directory because the input data
    # may not be grouped into swaths and/or may not be formatted as I need,
    # either internally or in terms of file naming convention.
    my (%options) = @_;
    my $name = (caller(0))[3];
    $name =~ s/main/$MAIN/g;
    my $sat = $options{sat};
    my @process_dir = ();
    mkpath('links') or die "$name\: WORK_DIR/links cannot be made.\nAbort";
    foreach (@{$options{input_files}}) {
      my $target = File::Spec -> catfile($options{input_dir}, $_);
      my $linkname = File::Spec -> catfile('links', $_);
      symlink($target, $linkname) or die "$name\: Cannot make link to $target\nAbort";
    }
    # ` ` ` ` ` ` ` ` ` ` ` ` ` ` ` ` ` ` ` ` ` ` ` ` ` ` ` ` ` ` ` ` ` ` `
    # JPSS JPSS JPSS JPSS JPSS JPSS JPSS JPSS JPSS JPSS JPSS JPSS JPSS JPSS
    # `````````````````````````````````````````````````````````````````````
    (grep {$_ eq $sat} qw/npp j01/) and do {
      my @fields = split(/\_/, basename(@{$options{input_files}}));
      my $origin = $fields[7];
      my $domain = (split(/\./, basename($fields[8])))[0];
      # If $options{break}, which is the default, we run file through nagg
      # We do this to be sure of presenting 32-second granules to NUCAPS
      if ($options{break}) {
        mkpath('input') or die "$name\: WORK_DIR/input cannot be made.\nAbort";
        # This OK because @{$options{input_files}} already globbed with correct mask!
        foreach my $sensor (qw/TATMS SCRIS SCRIF/) {
          my %nagg = ();
          my $globber = File::Spec -> catfile('links', "$sensor*.h5");
          foreach (sort(glob($globber))) {
            my $mask = substr(basename($_), 10, 34);
            ($sensor eq 'TATMS') and $mask = "GATMO_$sat\_$mask*.h5";
            ($sensor eq 'SCRIS') and $mask = "GCRSO_$sat\_$mask*.h5";
            ($sensor eq 'SCRIF') and $mask = "GCRSO_$sat\_$mask*.h5";
            my @geo = glob("links/$mask");
            ($#geo > -1) and $nagg{$_} = $geo[0];
          }
          foreach my $key (sort keys %nagg) {
            my $cmd = "$H5NAGG -S -t $sensor -n 1 -d input -O $origin -D $domain $key $nagg{$key}";
            ($options{verbose} < 3) and $cmd = $cmd." > /dev/null";
            system($cmd) == 0 or die "$name\: Could not run command $cmd\nAbort";
          }
        }
      } else {
        # Instead of loading directory 'input' with the 32-second granule outputs of
        # nagg opersating on the linked .h5 files in 'links', we assume those linked
        # files are already 32s and make the 'input' directory a link to them.
        symlink('links', 'input') or die "$name\: Cannot make link to links\nAbort";
        # NOTE: This may have repercussion for handling the VIIRS data, directly below,
        # but that has not yet been implemented so you can review when it is.
      }
      # If Clouds selected then I will have JRR-CloudHeight, JRR-CloudMask and GMTCO or
      # GMODO files in links/ that need to come across disaggregated to inputs/. I have no
      # expectation that nagg is going to work with JRR files, so I am going to have to
      # assume that they are already presented as 32s granules. The GMTCO file can be
      # disaggregated into input/ whether it needs to be or not.
      if (index($options{output}, 'C') != -1) {
        my %nagg = ();
        my $globber = File::Spec -> catfile('links', "GMTCO*.h5");
        foreach (sort(glob($globber))) {
          my $mask = substr(basename($_), 10, 34);
          $mask = "GMTCO_$sat\_$mask*.h5";
          my @geo = glob("links/$mask");
          ($#geo > -1) and $nagg{$_} = $geo[0];
        }
        foreach my $key (sort keys %nagg) {
          my $cmd = "$H5NAGG -g GMTCO -n 1 -d input -O $origin -D $domain $key $nagg{$key}";
          ($options{verbose} < 3) and $cmd = $cmd." > /dev/null";
          system($cmd) == 0 or die "$name\: Could not run command $cmd\nAbort";
        }
        $globber = File::Spec -> catfile('links', "JRR-Cloud*.nc");
        foreach (sort(glob($globber))) {
          copy($_, 'input') or do {
            ($options{verbose} >= 1) and warn "WARNING: Copy failed: $!\n"; };
        }
      }
      my $ncglobber = File::Spec -> catfile('input', 'JRR*.nc');
      my $globber = File::Spec -> catfile('input', '*.h5');
      my @input_files = ();
      foreach (sort(glob($globber))) { push(@input_files, basename($_)); }
      # I want to identify the granules which make up continuous swaths and have
      # them (or rather links to them) in separate, numbered swath directories under
      # the temporary working directory (which is currently, and by design, $PWD).
      my %swaths = granules_to_swaths_jpss(@input_files);
      ($options{verbose} > 2) and pretty_dump('%swaths', %swaths);
      my $iswath = 0;
      foreach my $key (sort keys %swaths) {
        my $granules = sprintf("swath_%03d", ++$iswath);
        mkpath($granules) or die "$name\: WORK_DIR $granules cannot be made.\nAbort";
        foreach (@{$swaths{$key}}) {
          my $target = File::Spec -> catfile('../input', $_);
          if (0==1) { # This cannot occur - just a repack file logic placeholder
            my $cmd = "$H5REPACK -n -f NONE $target $granules/$_";
            system($cmd) == 0 or die "$name\: Could not run command $cmd\nAbort";
          } else {
            my $linkname = File::Spec -> catfile($granules, $_);
            symlink($target, $linkname) or die "$name\: Cannot make link to $target\nAbort";
          }
        }
        # I am going to link every JRR file into every swath and sort out later
        # which one needs to be available to each specific granule
        foreach (sort(glob($ncglobber))) {
          my $target = File::Spec -> catfile('../input', basename($_));
          my $linkname = File::Spec -> catfile($granules, basename($_));
          symlink($target, $linkname) or die "$name\: Cannot make link to $target\nAbort";
        }
        # This is to check that I have nicely configured swath directories, meaning
        # that I have GATMO & TATMS files if mode requires microwave data, that the
        # GATMO & TATMS filecounts match.  Then do the same for GCRSO and SCRI* files.
        (grep {$_ eq $options{mode}} qw/HR_IRMW LR_IRMW/) and do {
          my @gatmo = sort(glob("$granules/GATMO*"));
          ($#gatmo > -1) or do {
            warn "WARNING: No GATMO found to process in /$granules\n";
            next; };
          my @tatms = sort(glob("$granules/TATMS*"));
          ($#tatms == $#gatmo) or do {
            warn "WARNING: Mis-match in TATMS/GATMO file count in /$granules\n";
            next; };
        };
        my @gcrso = sort(glob("$granules/GCRSO*"));
        ($#gcrso > -1) or do {
          warn "WARNING: No GCRSO found to process in /$granules\n";
          next; };
        my @scris = sort(glob("$granules/SCRI*"));
        ($#scris == $#gcrso) or do {
          warn "WARNING: Mis-match in SCRI/GCRSO file count in /$granules\n";
          next; };
        push(@process_dir, $granules);
      }
    };
    # ` ` ` ` ` ` ` ` ` ` ` ` ` ` ` ` ` ` ` ` ` ` ` ` ` ` ` ` ` ` ` ` ` ` ` `
    # MetOp MetOp MetOp MetOp MetOp MetOp MetOp MetOp MetOp MetOp MetOp MetOp
    # ```````````````````````````````````````````````````````````````````````
    (grep {$_ eq $sat} qw/m01 m02 m03/) and do {
      my ($mdrstart, $sens) = split(/\_/, $options{mode});
      if ($options{xfiles}) {
        # Use these files directly without renaming or disaggregating; make a link
        symlink('links', 'input') or die "$name\: Cannot make link to links\nAbort";
      } else {
        # Files have names that need to be changed, and IASI L1C must be disggregated
        mkpath('input') or die "$MAIN\:: WORK_DIR/input cannot be made.\nAbort";
        #
        # Convert the specified @options{input_files} to required filenames and formats
        # `````````````````````````````````````````````````````````````````````````````
        foreach (@{$options{input_files}}) {
          my $fullfile = File::Spec -> catfile($options{input_dir}, $_);
          # AMSUA DB - make sure big endian and NOAA name
          (index($_, 'amsua') < 0) or do {
            (length($_)==36) or next;
            my $outfile = make_L1B_bigEndian('amsua', $fullfile);
            my $inplace = 'NSS'.substr($outfile, 3);
            $inplace =~ s/X.MB.D/X.M1.D/g; $inplace =~ s/X.MA.D/X.M2.D/g;
            move($outfile, "input/$inplace") or warn "$MAIN\:: Could not move $outfile\n"; };
          # AMSUA NOAA - symlink
          (index($_, 'AMAX') < 0) or do {
            (length($_)==42) or next;
            symlink($fullfile, "input/$_") or warn "$MAIN\:: Could not symlink $_\n"; };
          # MHS DB - make sure big endian and NOAA name
          (index($_, 'mhs') < 0) or do {
            (length($_)==34) or next;
            my $outfile = make_L1B_bigEndian('mhs', $fullfile);
            my $inplace = 'NSS'.substr($outfile, 3);
            $inplace =~ s/X.MB.D/X.M1.D/g; $inplace =~ s/X.MA.D/X.M2.D/g;
            move($outfile, "input/$inplace") or  warn "$MAIN\:: Could not move $outfile\n"; };
          # MHS NOAA - symlink
          (index($_, 'MHSX') < 0) or do {
            (length($_)==42) or next;
            symlink($fullfile, "input/$_") or warn "$MAIN\:: Could not symlink $_\n"; };
          # IASI
          (index($_, 'IASI_xxx_1C') < 0) or do {
            if ($options{xfiles}) {
              # NOAA - symlink
              symlink($fullfile, "input/$_") or warn "$MAIN\:: Could not symlink $_\n";
            } else {
              # DB - slice up with kai then move
              my $kai_skip = compute_kai_skip($mdrstart, $fullfile);
              my $cmd = "$KAI --pdu XXX --skip $kai_skip --in $fullfile 2>/dev/null";
              my $dribble = `$cmd`;
              my ($n, $undersize) = (0, 1);
              foreach (split("\n",$dribble)) {
                (index($_, 'Setting MPHR TOTAL_MDR') < 0) or do {
                  my $mdr = trim_string((split(' to ', $_))[1]);
                  (grep {$_ eq $mdr} qw/22 23/) and $undersize = 0; };
                (index($_, 'Setting MPHR PRODUCT_NAME') < 0) or do {
                  my $granule = trim_string((split(' to ', $_))[1]);
                  my $infile = sprintf("XXX.%05d.pdu", $n++);
                  ($undersize) and do { unlink($infile); next };
                  $undersize = 1;
                  my $inplace = "input/$granule";
                  move($infile, $inplace) or warn "$MAIN\:: Could not move $infile\n";
                };
              }
            }
          };
        }
      }
      # What dates are covered by the files to process, just use IASI files
      # ```````````````````````````````````````````````````````````````````
      my $globber = File::Spec -> catfile('input', '{IASI_xxx_1C,NSS.}*');
      my @input_files = ();
      foreach (sort(glob($globber))) { push(@input_files, basename($_)); }
      my %swaths = granules_to_swaths_metop(@input_files);
      ($options{verbose} > 2) and pretty_dump('%swaths', %swaths);
      my $iswath = 0;
      foreach my $key (sort keys %swaths) {
        my $granules = sprintf("swath_%03d", ++$iswath);
        mkpath($granules) or die "$name\: WORK_DIR $granules cannot be made.\nAbort";
        foreach (@{$swaths{$key}}) {
          my $target = File::Spec -> catfile('../input', $_);
          my $linkname = File::Spec -> catfile($granules, $_);
          symlink($target, $linkname) or die "$name\: Cannot make link to $target\nAbort";
        }
        push(@process_dir, $granules);
      }
    };
    @{$options{process_dir}} = @process_dir;
    return %options;
  }

# -------------------
  sub get_commandline
# ===================
# %options = get_commandline($revision, $pwd);
  {
    # If things go bad in here, jump to usage() which terminates with exit(9).
    # Note that I had to hack in the Python-like -vv & -vvv support.
    # This routine assumes standard module Getopt::Long has been loaded.
    # abs_path() might be dangerous for the purpose of setting the work (and,
    # equivalently, the output) directory - see this post:
    # http://bugs.debian.org/cgi-bin/bugreport.cgi?bug=257568
    my ($rev, $pwd) = (shift, shift);
    my $name = (caller(0))[3];
    $name =~ s/main/$MAIN/g;
    my $codeset = langinfo(CODESET);
    @ARGV = map{decode $codeset, $_} @ARGV;
    foreach (@ARGV) { s/\?/-\?/g; }
    my %options = ();
    GetOptions(\%options,
      'version|V',
      'help|?',
      'nothing|n',
      'input_files|i=s',
      'sat|s=s',
      'work|w=s',
      'output|o=s',
      'mode|m=s',
      'dynanc|d=s',
      'gfs|g=i',
      'fovs|f=i',
      'processors|p=i',
      'verbose|v+', 'vv', 'vvv',
      'xfiles|x!',
      'aggregate|a!',
      'break|b!',
      'quarantine|q!',
      'clean|c!') or usage("\tERROR --- $name\: Unknown command option.\n\n");
    defined($options{version}) and die "$rev\n";
    defined($options{help}) and usage();
    $options{nothing} = defined($options{nothing}) ? $options{nothing} : 0;
    defined($options{input_files}) or 
      usage("\tERROR --- $name\: Required INPUT_FILES missing.\n\n");
    # Checks on MISSIONS and MODE
    defined($options{sat}) or
      usage("\tERROR --- $name\: Required MISSION is missing.\n\n");
    (grep {$_ eq $options{sat}} qw/npp j01 m01 m02 m03/) or
      usage("\tERROR --- $name\: Unsupported MISSION $options{sat}.\n\n");
    (grep {$_ eq $options{sat}} qw/npp j01/) and do {
      $options{mode} = defined($options{mode}) ? $options{mode} : 'HR_IRMW';
      $options{mode} = uc($options{mode});
      (grep {$_ eq $options{mode}} qw/HR_IRMW LR_IRMW HR_IRO LR_IRO/) or
        usage("\tERROR --- $name\: Unsupported MODE $options{mode}.\n\n"); };
    (grep {$_ eq $options{sat}} qw/m01 m02 m03/) and do {
      $options{mode} = defined($options{mode}) ? $options{mode} : 'M_IRMW';
      $options{mode} = uc($options{mode});
      (grep {$_ eq $options{mode}} qw/IRMW IRO/) and do {
        $options{mode} = sprintf("M_%s", $options{mode}); };
      (grep {$_ eq $options{mode}} qw/S_IRMW M_IRMW E_IRMW S_IRO M_IRO E_IRO/) or
        usage("\tERROR --- $name\: Unsupported MODE $options{mode}.\n\n"); };
    (defined($options{work})) or $options{work} = $pwd;
    $options{output} = defined($options{output}) ? $options{output} : 'R';
    $options{output} = uc($options{output});
    ($options{output} eq 'P') or do { # Hidden option Preprocessor only
      my $chk = join('', sort split(//, $options{output}));
      my @valids = qw/CORS COR COS CRS CO CR CS C ORS OR OS O RS R S/;
      (grep {$_ eq $chk} @valids) or
        usage("\tERROR --- $name\: Unsupported OUTPUT string $options{output}\n\n");
      if (index($options{output}, 'P') == -1) {
        $options{output} = join('', 'P', $options{output});
      }
    };
    # If we have C $options{output} for MetOp, remove - it is unsupported
    (grep {$_ eq $options{sat}} qw/m01 m02 m03/) and do {
      if (index($options{output}, 'C') != -1) {
        $options{output} = join('', split(/C/, $options{output}));
      }
    };
    (defined($options{gfs})) or $options{gfs} = 3;
    (defined($options{fovs})) or do { 
      ($options{sat} eq 'npp') and $options{fovs} = 7;
      ($options{sat} eq 'j01') and $options{fovs} = 9; };
    (defined($options{dynanc})) or 
      $options{dynanc} = File::Spec -> catfile($ENV{CSPP_DYNAMIC_ANCIL_DIR});
    foreach (qw/dynanc work/) {
      $options{$_} = File::Spec -> catfile(abs_path(dirname($options{$_})),
        basename($options{$_}));
    }
    (defined($options{processors})) or $options{processors} = 1;
    $options{verbose} = defined($options{verbose}) ? $options{verbose} : 0;
    $options{vv} = defined($options{vv}) ? $options{vv} : 0;
    $options{vvv} = defined($options{vvv}) ? $options{vvv} : 0;
    $options{verbose} = $options{verbose} + 2 * $options{vv};
    $options{verbose} = $options{verbose} + 3 * $options{vvv};
    $options{verbose} = min($options{verbose}, 3);
    delete @options{'vv', 'vvv'};
    $options{clean} = defined($options{clean}) ? $options{clean} : 1;
    $options{aggregate} = defined($options{aggregate}) ? $options{aggregate} : 1;
    $options{break} = defined($options{break}) ? $options{break} : 1;
    $options{xfiles} = defined($options{xfiles}) ? $options{xfiles} : 0;
    $options{quarantine} = defined($options{quarantine}) ? $options{quarantine} : 0;
    ($options{gfs} > 1 and $options{gfs} < 3) and $options{quarantine} = 1; # no LFTP fetch.
    return %options;
  }

# ----------------------------
  sub get_datetime_from_tokens
# ============================
# ($YYYY, $MM, $DD, $hh, $mm, $ss) = get_datetime_from_tokens($string, $tokens);
  {
    # $string is probably a filename with date/time elements you want to extract.
    # $tokens is a matchin string of date/time tokens matched to those elements.
    # See function token_value() for more information.
    my ($string, $tokens) = (shift, shift);
    my ($YYYY, $MM, $DD, $hh, $mm, $ss) = undef;
    $YYYY = token_value('%Y', $string, $tokens);
    $MM = token_value('%m', $string, $tokens);
    $DD = token_value('%d', $string, $tokens);
    $hh = token_value('%H', $string, $tokens);
    $mm = token_value('%M', $string, $tokens);
    $ss = token_value('%S', $string, $tokens);
    return ($YYYY, $MM, $DD, $hh, $mm, $ss);
  }

# ---------------------------
  sub gfs_fetch_with_fallback
# ===========================
# ($avn1, $avn2) = gfs_fetch_with_fallback($beg, \%dynanc_loc, %options)
  {
    # $beg is the start time of the CrIS SDR or IASI L1C file
    # %dynanc_loc is all the location ifromation, local and remote
    # %options the command line options + more controls
    # The "fallback" is actually coded into function heap_avn_files().
    my $beg = shift;
    my %dynanc_loc = %{(shift)};
    my (%options) = @_;
    my $name = (caller(0))[3];
    $name =~ s/main/$MAIN/g;
    my $waitawhile = $SLEEP_GFS_MINUTES; # global variable
    #
    # Use beginning time of CrIS SDR or IASI L1C to define the AVN gfs files
    # ``````````````````````````````````````````````````````````````````````
    my %avn = heap_avn_files($beg, %dynanc_loc);
    ($options{verbose} == 3) and pretty_dump('%avn', %avn);
    #
    # Final check is that we have available the GFS data required
    # ```````````````````````````````````````````````````````````
    my ($url, $local, $avn1, $avn2) = ('N/A', 'N/A', 'N/A', 'N/A');
    foreach (qw/table spare/) {
      ($url, $local, $avn1, $avn2) = @{$avn{$_}};
      foreach my $avn (($avn1, $avn2)) {
        my $target = File::Spec -> catfile($local, $avn);
        -e $avn or do {
          lftp_mirror($url, $avn, $local, %options);
          -e $target and symlink($target, $avn) or
            warn "$name\: WARNING: Cannot link to $target\n";
          ($options{gfs}==1) and next; # don't need multiple attempts at local repo
          ($options{gfs}==2) and next; # don't need multiple attempts at local repo
          ($options{quarantine}) and next; # don't need multiple attempts at local repo
          -e $avn and next; # we have $target and link, so we are done for this $avn
          my $j = $waitawhile; # so $j is still in scope after the following loop
          for ($j=$waitawhile; $j>0; $j--) {
            ($options{verbose} >= 1) and warn "$name\: SLEEPING at minute $j waiting for $avn\n";
            sleep 60;
            lftp_mirror($url, $avn, $local, %options);
            -e $target or next;
            symlink($target, $avn) or warn "$name\: WARNING: Cannot link to $target\n";
            last;
          }
          $waitawhile = $j;
        };
      }
      ((-e $avn1) and (-e $avn2)) and last;
    }
    return ($avn1, $avn2);
  }

# ---------------------------
  sub granules_to_swaths_jpss
# ===========================
# %swaths = granules_to_swaths_jpss(@files);
  {
    # @files are input CrIS and ATMS files, and NOT including GMTCO/JRR files.
    # %swaths are key:value pairs $timegm:@filename_list
    # Break list of files into groups that comprise contiguous overpass swaths
    # to assist cspp_preprocess() in staging data. This is not a complicated
    # routine - all it does is group files together that fall within a time
    # window and ASSUMES them to belong to the same overpass; a reasonable
    # assumption for DB users, but details like dropped granules are wholly
    # ignored. The times are extracted from the well-specified NPP filenaming
    # convention.  ATMS granule-to-granule contiguity will need to be checked
    # later once you have built candidate CrIS granules to process and chosen
    # the fore-, mid- and aft ATMS granules.
    my @files = @_;
    my $name = (caller(0))[3];
    $name =~ s/main/$MAIN/g;
    my $time_window = 60 * $AGG_WINDOW_MINUTES;
    my $atms_beg = 2 * 60; # 2 mins of ATMS overlap
    my $atms_end = $time_window + $atms_beg;
    my %swaths = ();
    my ($YYYY, $MM, $DD, $hh, $mm, $ss, $found) = undef;
    #
    # First loop distributes CrIS files uniquely to swaths
    # ````````````````````````````````````````````````````
    my $cris = 'SCRIx_xxx_d%Y%m%d_t%H%M%Sx'; # start of filename is enough
    foreach (sort @files) {
      (substr($_, 1, 3) eq 'ATM') and next;
      ($YYYY, $MM, $DD, $hh, $mm, $ss) = get_datetime_from_tokens($_, $cris);
      is_plausible_datetime($YYYY, $MM, $DD, $hh, $mm, $ss) or
        die "$name\: Date from JPSS CrIS file $_ implausible\nAbort";
      my $t = timegm_nocheck($ss, $mm, $hh, $DD, $MM-1, $YYYY);
      $found = 0;
      foreach my $key (sort keys %swaths) {
        (abs($key - $t) < $time_window) and do {
          push(@{$swaths{$key}}, $_);
          $found = 1; last; };
      }
      ($found) or @{$swaths{$t}} = ($_);
    }
    #
    # Second loop distributes ATMS non-uniquely to swaths
    # ```````````````````````````````````````````````````
    my $atms = 'TATMS_xxx_d%Y%m%d_t%H%M%Sx'; # start of filename is enough
    foreach (sort @files) {
      (substr($_, 1, 3) ne 'ATM') and next;
      ($YYYY, $MM, $DD, $hh, $mm, $ss) = get_datetime_from_tokens($_, $atms);
      is_plausible_datetime($YYYY, $MM, $DD, $hh, $mm, $ss) or
        die "$name\: Date from JPSS ATMS file $_ implausible\nAbort";
      my $t = timegm_nocheck($ss, $mm, $hh, $DD, $MM-1, $YYYY);
      foreach my $key (sort keys %swaths) {
        (($key - $t) > $atms_beg) and next;
        (($t - $key) > $atms_end) and next;
        push(@{$swaths{$key}}, $_);
      }
    }
    return %swaths;
  }

# ----------------------------
  sub granules_to_swaths_metop
# ============================
# %swaths = granules_to_swaths_metop(@files);
  {
    # %swaths are key:value pairs $timegm:@filename_list
    # Break list of files into groups that comprise contiguous overpass swaths
    # to assist cspp_preprocess() in staging data. This is not a complicated
    # routine - all it does is group files together that fall within a time
    # window and ASSUMES them to belong to the same overpass; a reasonable
    # assumption for DB users, but details like dropped granules are wholly
    # ignored. Times are extracted from the well-specified IASI L1C filenaming
    # convention.  AMSUA/MHS granule-to-granule contiguity will need to be
    # checked later once you have built candidate IASI granules to process.
    my @files = @_;
    my $name = (caller(0))[3];
    $name =~ s/main/$MAIN/g;
    my $time_window = 60 * $AGG_WINDOW_MINUTES;
    my %swaths = ();
    my ($YYYY, $MM, $DD, $DOY, $hh, $mm, $ss, $tod, $found) = undef;
    #
    # First loop distributes IASI L1C files uniquely to swaths
    # ````````````````````````````````````````````````````````
    my $iasi = 'IASI_xxx_1C_M0x_%Y%m%d%H%M%SZ'; # start of filename is enough
    foreach (sort @files) {
      (substr($_, 0, 4) ne 'IASI') and next;
      ($YYYY, $MM, $DD, $hh, $mm, $ss) = get_datetime_from_tokens($_, $iasi);
      is_plausible_datetime($YYYY, $MM, $DD, $hh, $mm, $ss) or
        die "$name\: Date from MetOp IASI L1C file $_ implausible\nAbort";
      my $t = timegm_nocheck($ss, $mm, $hh, $DD, $MM-1, $YYYY);
      $found = 0;
      foreach my $key (sort keys %swaths) {
        (abs($key - $t) < $time_window) and do {
          push(@{$swaths{$key}}, $_);
          $found = 1; last; };
      }
      ($found) or @{$swaths{$t}} = ($_);
    }
    #
    # Second loop finds AMSUA/MHS files that overlap IASI L1C files in swaths
    # ```````````````````````````````````````````````````````````````````````
    my $nss1 = 'NSS.xxxX.Mx.D%y%j.S%H%M.E1444';
    my $nss2 = 'NSS.xxxX.Mx.D%y%j.Sxxxx.E%H%M';
    foreach my $key (sort keys %swaths) {
      my $l = @{$swaths{$key}}[-1];          # The last file in the swath.
      ($YYYY, $MM, $DD, $hh, $mm, $ss) = get_datetime_from_tokens($l, $iasi);
      is_plausible_datetime($YYYY, $MM, $DD, $hh, $mm, $ss) or
        die "$name\: DateTime from MetOp IASI L1C file $_ implausible\nAbort";
      my $lst = timegm_nocheck($ss, $mm, $hh, $DD, $MM-1, $YYYY);
      foreach (sort @files) {
        (substr($_, 0, 4) ne 'NSS.') and next;
        $YYYY = '20'.token_value('%y', $_, $nss1);
        $DOY = token_value('%j', $_, $nss1);
        $hh = token_value('%H', $_, $nss1);
        $mm = token_value('%M', $_, $nss1);
        is_plausible_doytime($YYYY, $DOY, $hh, $mm) or
          die "$name\: DoyTime from MetOp NSS. file $_ implausible\nAbort";
        $tod = $hh.$mm; # time of day as hhmm
        my $beg = timegm_nocheck(0, $mm, $hh, $DOY, 0, $YYYY);
        $hh = token_value('%H', $_, $nss2);
        $mm = token_value('%M', $_, $nss2);
        is_plausible_doytime($YYYY, $DOY, $hh, $mm) or
          die "$name\: DoyTime from MetOp NSS. file $_ implausible\nAbort";
        ($tod gt $hh.$mm) and $DOY++;
        my $end = timegm_nocheck(0, $mm, $hh, $DOY, 0, $YYYY);
        ($end < $key) and next;
        ($beg > $lst) and next;
        push(@{$swaths{$key}}, $_);
      }
    }
    return %swaths;
  }

# ------------------
  sub heap_avn_files
# ==================
# %avn = heap_avn_files($gmtbeg, %dynanc_loc);
  {
    # $gmtbeg is the beginning time of the CrIS SDR
    # %dynanc_loc is hash of dynamic ancillary data locations, local & remote
    # %avn is a hash of URL, LOCAL repo directory plus AVN file pairs, keyed
    # 'table' and 'spare', where 'table' is original AVN truth table as defined
    # by Chris Barnet, and 'spare' is the fallback possition if 'table' options
    # are unavailable. This fact cannot be known now, only when we go to the
    # ancillary data server or, if --quarantine is invoked, when we check the
    # local repo. Because the date is in both the URL and the LOCAL directory
    # name, and this could be different in the fallback positions, we have to
    # build all of $url, $local, $avn1 and $avn2 and return them in the hash;
    # I am going to just put then in 4 element vector in that order. The
    # bounding gfs files to use are for NUCAPS processing of CrIS SDR with
    # input $gmtbeg.
    my $gmtbeg = shift;
    my (%dynanc_loc) = @_;
    my $name = (caller(0))[3];
    $name =~ s/main/$MAIN/g;
    # AVN truth table as defined by Chris Barnet
    my %avn_truth_table = ('00z' => [('18z', 'F06', 'F09', 'yesterday')],
                           '03z' => [('00z', 'F03', 'F06', 'today')],
                           '06z' => [('00z', 'F06', 'F09', 'today')],
                           '09z' => [('06z', 'F03', 'F06', 'today')],
                           '12z' => [('06z', 'F06', 'F09', 'today')],
                           '15z' => [('12z', 'F03', 'F06', 'today')],
                           '18z' => [('12z', 'F06', 'F09', 'today')],
                           '21z' => [('18z', 'F03', 'F06', 'today')]);
    # AVN truth table fallback position 
    my %avn_truth_spare = ('00z' => [('12z', 'F12', 'F15', 'yesterday')],
                           '03z' => [('18z', 'F09', 'F12', 'yesterday')],
                           '06z' => [('18z', 'F12', 'F15', 'yesterday')],
                           '09z' => [('00z', 'F09', 'F12', 'today')],
                           '12z' => [('00z', 'F12', 'F15', 'today')],
                           '15z' => [('06z', 'F09', 'F12', 'today')],
                           '18z' => [('06z', 'F12', 'F15', 'today')],
                           '21z' => [('12z', 'F09', 'F12', 'today')]);
    gmtime($gmtbeg);
    my $YYYY = sprintf("%04d", $tm_year + 1900);
    my $ddd = sprintf("%03d", $tm_yday + 1);
    my $hh = sprintf("%02d", $tm_hour);
    my $key = sprintf("%02dz" , $hh - $hh % 3);
    my %avn = ();
    foreach (qw/table spare/) {
      my ($anal, $fct1, $fct2, $relday) = @{$avn_truth_table{$key}};
      /spare/ and ($anal, $fct1, $fct2, $relday) = @{$avn_truth_spare{$key}};
      my $avn1 = $dynanc_loc{GFSTEMPLATE};
      my $avn2 = $dynanc_loc{GFSTEMPLATE};
      $hh = substr($anal, 0, 2);
      my ($ff1, $ff2) = (substr($fct1, 1, 2), substr($fct2, 1, 2));
      $avn1 =~ s/%f/$ff1/g;
      $avn2 =~ s/%f/$ff2/g;
      ($relday eq 'yesterday') and $ddd--;
      gmtime(timegm_nocheck(0, 0, $hh, $ddd, 0, $YYYY));
      $YYYY = sprintf("%04d", $tm_year + 1900);
      my $MM = sprintf("%02d", $tm_mon + 1);
      my $DD = sprintf("%02d", $tm_mday);
      $ddd = sprintf("%03d", $tm_yday + 1);
      $hh = sprintf("%02d", $tm_hour);
      my $YY = substr($YYYY, 2 , 2);
      my @tokens = qw/ %Y   %y   %m    %j   %d   %H/;
      my @values = ($YYYY, $YY, $MM, $ddd, $DD, $hh);
      my $site = $dynanc_loc{'GFSSITE'};
      my $proto = $dynanc_loc{'GFSPROTO'};
      my $local = $dynanc_loc{'GFSLOCAL'};
      my $remote = $dynanc_loc{'GFSREMOTE'};
      for (my $j=0; $j<=$#tokens; $j++) {
        $local =~ s/$tokens[$j]/$values[$j]/g;
        $remote =~ s/$tokens[$j]/$values[$j]/g; 
        $avn1 =~ s/$tokens[$j]/$values[$j]/g; 
        $avn2 =~ s/$tokens[$j]/$values[$j]/g; 
      }
      # ********** MAKE PATH SO THAT WARNINGS CAN BE REPORTED **********
      if (! -d $local) {
        eval { mkpath($local) };
        if ($@) { warn "$name\: WARNING $local cannot be made\n"; }
      }
      my $url = "$proto$site/$remote";
      @{$avn{$_}} = ($url, $local, $avn1, $avn2);
    }
    return (%avn);
  }

# -----------------------------------------------------------
  sub is_integer { defined $_[0] and $_[0] =~ /^[+-]?\d+$/; }
# ===========================================================

# -------------------------
  sub is_plausible_datetime
# =========================
# $status = is_plausible_datetime($YYYY, $MM, $DD, $hh, $mm, $ss);
  {
    # NOTE: Year < 2000 is considered implausible in this function.
    my ($YYYY, $MM, $DD, $hh, $mm, $ss) = (@_);
    my $implausible_year = 2000;
    is_integer($YYYY) or return 0;
    $YYYY < $implausible_year and return 0;
    is_integer($MM) or return 0;
    $MM >= 1 and $MM <= 12 or return 0;
    is_integer($DD) or return 0;
    $DD >= 1 and $DD <= 31 or return 0;
    is_integer($hh) or return 0;
    $hh >= 0 and $hh <= 23 or return 0;
    is_integer($mm) or return 0;
    $mm >= 0 and $mm <= 59 or return 0;
    is_integer($ss) or return 0;
    $ss >= 0 and $ss <= 59 or return 0;
    return 1;
  }

# ------------------------
  sub is_plausible_doytime
# ========================
# $status = is_plausible_doytime($YYYY, $DOY, $hh, $mm);
  {
    # NOTE: Year < 2000 is considered implausible in this function.
    my ($YYYY, $DOY, $hh, $mm) = (@_);
    my $implausible_year = 2000;
    is_integer($YYYY) or return 0;
    $YYYY < $implausible_year and return 0;
    is_integer($DOY) or return 0;
    $DOY >= 1 and $DOY <= 366 or return 0;
    is_integer($hh) or return 0;
    $hh >= 0 and $hh <= 23 or return 0;
    is_integer($mm) or return 0;
    $mm >= 0 and $mm <= 59 or return 0;
    return 1;
  }

# -------------------------------
  sub job_coverage_start_end_jpss
# ===============================
# ($jcs, $jce) = job_coverage_start_end_jpss($cris_geo);
  {
    # Compute the job coverage start and end strings based on CrIS GEO
    # granule name, $cris_geo.  The problem I encountered is that this
    # name includes the YearMonthDay only for the start time, so that if
    # end time is in next day, just prepending the YearMonthDay string will
    # be wrong.  So I need to check if end time (Hour,Min,Sec) precedes start
    # time, and if so, add a day using some Perl date arithmetic (because it
    # could be end of month or end of year, and I need to roll-over correctly).
    my $cris_geo = shift;
    my $tag = substr($cris_geo, 6, 31);
    my $sdate = substr($tag, 5, 8);
    my $edate = $sdate;
    my $stime = substr($tag, 15, 7);
    my $etime = substr($tag, 24, 7);
    if ($etime < $stime) {
      my $YYYY = substr($sdate, 0, 4);
      my $MM = substr($sdate, 4, 2);
      my $DD = substr($sdate, 6, 2);
      # $MM-1 because month is zero-based; $DD+1 because we want next day.
      gmtime(timegm_nocheck(0, 0, 0, $DD+1, $MM-1, $YYYY));
      $YYYY = sprintf("%04d", $tm_year + 1900);
      $MM = sprintf("%02d", $tm_mon + 1);
      $DD = sprintf("%02d", $tm_mday);
      $edate = $YYYY.$MM.$DD;
    }
    my $jcs = $sdate.$stime;
    my $jce = $edate.$etime;
    return ($jcs, $jce);
  }

# ---------------
  sub lftp_mirror
# ===============
# $status = lftp_mirror($site, $file, $local, %options);
  {
    # executes lftp in mirror mode, but if NOT $options{quarantine}
    my ($site, $file, $local) = (shift, shift, shift);  
    my (%options) = @_;
    my $name = (caller(0))[3];
    $name =~ s/main/$MAIN/g;
    my $v = '';
    ($options{verbose} == 3) and $v = '-v';
    -e "$local/$file" and do {
      ($options{verbose} == 3) and print "$name\: $file FOUND in local repo\n";
      return 1;
    };
    ($options{quarantine}) and do {
      ($options{verbose} >= 1) and 
        warn "$name\: WARNING: $file NOT FOUND in local repo AND --quarantine\n";
      return 0;
    };
    (-d $local) or do {
      ($options{verbose} >=2) and print "Making $local\n";
      mkpath($local) or
        die "$name\: $local cannot be made.\nAbort"; };
    open(my $fh, '>script_file.lftp') or die "$name\: Cannot open script_file.lftp $!";
    print $fh "set dns:order inet inet6\n";            # default is 'inet6 inet'
    #print $fh "set ssl:verify-certificate no\n";       # default is ‘yes'
    #print $fh "set net:timeout 1\n";                   # default is '5'
    #print $fh "set net:max-retries 5\n";               # default is '10'
    #print $fh "set net:persist-retries 1\n";           # default is '5'
    #print $fh "set net:reconnect-interval-base 6\n";   # default is '60'
    print $fh "mirror $v -L -n --no-empty-dirs $site -I $file $local\n";
    close($fh);
    my $cmd = "flock $local $LFTP -f script_file.lftp";
    system($cmd) == 0 or do {
      ($options{verbose} >= 1) and warn "$name\: WARNING: $cmd failed\n";
      return 0; };
    return 1;
  }

# --------------------------
  sub make_heap_pcf_template
# ==========================
# make_heap_pcf_template($step, %options);
  {
    # $step is a letter from $options{output} indicating PCF template to make
    # <stuff in angle brackets> is to be replaced in runtime PCF
    my $step = shift;
    my (%options) = @_;
    my $name = (caller(0))[3];
    $name =~ s/main/$MAIN/g;
    #
    # No Clouds support for MetOp
    # ```````````````````````````
    (($step eq 'C') and (grep {$_ eq $options{sat}} qw/m01 m02 m03/)) and return;
    my $pcf = 'HEAP_XXX.PCF.template'; # $pcf is written in $PWD==$tmpdir
    my $longstep ='';
    ($step eq 'P') and $longstep ='Preprocessor';
    ($step eq 'C') and $longstep ='Clouds';
    ($step eq 'O') and $longstep ='OLR';
    ($step eq 'S') and $longstep ='Subsetter';
    ($step eq 'R') and $longstep ='Retrieval';
    $pcf =~ s/XXX/$longstep/g;
    ($options{verbose} >= 1) and print "Making template $pcf\n";
    open(my $fh, ">$pcf") or die "$name\: Cannot open template $pcf for write.\nAbort";
    my ($res, $sens) = split(/\_/, $options{mode});
    # 
    # Common to all PCFs
    # ``````````````````
    print $fh "BIN_DIR=$OPS/Common_Bin\n";
    print $fh "EMAIL_ERROR_FLAG=NO\n";
    print $fh "MACHINE=nomachine.nowhere\n";
    print $fh "MAIL=nobody\@nowhere\n";
    print $fh "WORKING_DIR=.\n";
    print $fh "LOG_FILE=HEAP_$longstep.log\n";
    print $fh "PSF_FILE=HEAP_$longstep.PSF\n";
    print $fh "PERL_LOC=/usr/bin/perl\n";
    print $fh "SATELLITE=$options{sat}\n";
    print $fh "SCRIPT_DIR=$OPS/scripts\n";
    print $fh "STATIC_ANCILLARY_DIR=$OPS/ancillary\n";
    print $fh "H5DUMP_LOC=$H5DUMP\n";
    # Not actually common to ALL, but safe I think
    print $fh "UUIDGEN_LOC=$UUIDGEN\n";
    print $fh "TIME_OUT=\n";
    print $fh "production_environment=DEV\n";
    print $fh "production_site=CSPP\n";
    # 
    # Common to all PCFs but dependent on mission
    # ```````````````````````````````````````````
    (grep {$_ eq $options{sat}} qw/j01 npp/) and do {
      print $fh "CDL_TEMPLATE_DIR=$OPS/CDLFILES/CrIS\n"; };
    (grep {$_ eq $options{sat}} qw/m01 m02 m03/) and do {
      print $fh "CDL_TEMPLATE_DIR=$OPS/CDLFILES/IASI\n"; };
    #
    # All except Preprocessor
    # ```````````````````````
    (grep {$_ eq $step} qw/C O S R/) and do {
      print $fh "job_coverage_end=<JCE>\n";
      print $fh "job_coverage_start=<JCS>\n";
      print $fh "GRANULE_INFO_FILE=<GRANULE_INFO_FILE>\n";
      print $fh "IR_NC_FILE=<IR_NC_FILE>\n";
    };
    #
    # Preprocessor
    # ````````````
    ($step eq 'P') and do {
      print $fh "APOD_FLAG=YES\n";
      print $fh "IR_FILE=<IR_FILE>\n";
      print $fh "IR_GEO_FILE=<IR_GEO_FILE>\n";
      print $fh "USE_CLOUD_DATA_FLAG=<USE_CLOUD_DATA_FLAG>\n";
      ($sens eq 'IRO') and print $fh "IR_ONLY_FLAG=YES\n";
      ($sens eq 'IRMW') and print $fh "IR_ONLY_FLAG=NO\n";
      (grep {$_ eq $options{sat}} qw/j01 npp/) and do {
        print $fh "MW_FILE_1=<MW_FILE_1>\n";
        print $fh "MW_FILE_2=<MW_FILE_2>\n";
        print $fh "MW_FILE_3=<MW_FILE_3>\n";
        print $fh "MW_GEO_FILE_1=<MW_GEO_FILE_1>\n";
        print $fh "MW_GEO_FILE_2=<MW_GEO_FILE_2>\n";
        print $fh "MW_GEO_FILE_3=<MW_GEO_FILE_3>\n"; };
      (grep {$_ eq $options{sat}} qw/m01 m02 m03/) and do {
        print $fh "MW_FILE_1=<MW_FILE_1>\n";
        print $fh "MW_FILE_2=<MW_FILE_2>\n";
        print $fh "MW_FILE_3=<MW_FILE_3>\n";
        print $fh "MW_FILE_4=<MW_FILE_4>\n"; };
    };
    #
    # Clouds (JPSS only)
    # ``````````````````
    ($step eq 'C') and do {
      print $fh "CLOUD_FILE_1=<CLOUD_FILE_1>\n";
      print $fh "CLOUD_FILE_2=<CLOUD_FILE_2>\n";
      print $fh "CLOUD_FILE_3=<CLOUD_FILE_3>\n";
      print $fh "CLOUD_FILE_4=<CLOUD_FILE_4>\n";
      print $fh "CLOUD_FILE_5=<CLOUD_FILE_5>\n";
      print $fh "CLOUD_FILE_6=<CLOUD_FILE_6>\n";
      print $fh "CLOUD_GEO_FILE_1=<CLOUD_GEO_FILE_1>\n";
      print $fh "CLOUD_GEO_FILE_2=<CLOUD_GEO_FILE_2>\n";
      print $fh "CLOUD_GEO_FILE_3=<CLOUD_GEO_FILE_3>\n";
      print $fh "CLOUD_PADDING=14_seconds\n";
      print $fh "IR_GEO_FILE=<IR_GEO_FILE>\n";
    };
    #
    # Outgoing Longwave Radiation (OLR)
    # `````````````````````````````````
    ($step eq 'O') and do {
      # Covered by 'Common' and 'All except Preprocessor'
    };
    #
    # Subsetter
    # `````````
    ($step eq 'S') and do {
      # JPSS and MetOp are the same
      print $fh "IR_FILE=<IR_FILE>\n";
      print $fh "RUN_FULL_SUBSET_FLAG=YES\n";
      print $fh "RUN_PC_SUBSET_FLAG=\n";
    };
    #
    # Retrieval
    # `````````
    ($step eq 'R') and do {
      print $fh "BYTESWAP_FLAG=YES\n";
      print $fh "GFS_FILE_1=<GFS_FILE_1>\n";
      print $fh "GFS_FILE_2=<GFS_FILE_2>\n";
      print $fh "USE_CLOUD_DATA_FLAG=<USE_CLOUD_DATA_FLAG>\n";
      ($sens eq 'IRO') and print $fh "IR_ONLY_FLAG=YES\n";
      ($sens eq 'IRMW') and print $fh "IR_ONLY_FLAG=NO\n";
      if ($options{gfs} == 3) {
        print $fh "GFS_GRID_SIZE=0.5\n"
      } else {
        print $fh "GFS_GRID_SIZE=1.0\n"
      }
      if ($options{gfs} == 1) {
        print $fh "WGRIB_LOC=$WGRIB\n";
      } else {
        print $fh "WGRIB_LOC=$WGRIB2\n";
      }
      # JPSS specific
      (grep {$_ eq $options{sat}} qw/j01 npp/) and do {
        print $fh "NUM_FOVS=$options{fovs}\n";
      };
      # MetOp specific
      (grep {$_ eq $options{sat}} qw/m01 m02 m03/) and do {
        print $fh "AMSUA_NC_FILE=<AMSUA_NC_FILE>\n";
        print $fh "LAC_FLAG=LAC_ON\n";
        print $fh "MHS_NC_FILE=<MHS_NC_FILE>\n";
      };
    };
    close($fh) or die "$name\: Cannot close template $pcf.\nAbort";
    #
    # I'd rather have the PCF sorted - makes it easier to check
    # `````````````````````````````````````````````````````````
    open($fh, "<$pcf"); my @lines = <$fh>; close($fh);
    open($fh, ">$pcf"); print $fh sort(@lines); close($fh);
    return;
  }

# ----------------------
  sub make_L1B_bigEndian
# ======================
# $outfile = make_L1B_bigEndian($instrument, $infile);
  {
    # $instrument is one of 'amsua' or 'mhs'
    # $infile is the name of the L1B file to test/convert endian
    # $outfile is the big endian output filename created
    #
    # $outfile is extracted from the header of $infile, bytes 23-64 inclusive
    # (42 bytes). The reason for creating $outfile in WORK_DIR/tempdir is so
    # that we don't clobber data files in WORK_DIR if the user happens to
    # specify WORK_DIR as the place where the original L1B files are stored.
    # For files named like "amsual1b_noaa18_20131014_2145_43295.l1b" no problem,
    # but for files named like "NSS.AMAX.NN.D06032.S0122.E0317.B0361819.WI",
    # problem since this is the same as the filename stored in its header!
    #
    # NOTE: We coalesce repeated character variables to strings in the $Lhdr and
    # $Lrec buffer specification (because we are not actually applying any logic
    # to characters), but only after 'a42a8' in $Lhdr, since the a42 field is to
    # be used to define $outfile.
    #
    my ($instrument, $infile) = (shift, shift);
    my $name = (caller(0))[3];
    $name =~ s/main/$MAIN/g;
    my ($Lhdr, $Bhdr, $Lrec, $Brec, $count, $buf) = undef;
    my ($reclen, $year, $iyear, $doy, $idoy, $iout) = undef;
    ($instrument eq 'amsua') and do {
      ($reclen, $iout, $iyear, $idoy) = (2560, 11, 22, 23);
      $Lhdr = 'a3av9a42a8v4Vv2V2v2Vv3v5Vv2V2v2Vv14Va16V4v106V115a8v8V363v6Vv2V46';
      $Lrec = 'v4Vv2V4v16V95v6V2v94V61v620Vv2V2a28Vv156Vv2V2a32v6V'; };
    ($instrument eq 'mhs') and do {
      ($reclen, $iout, $iyear, $idoy) = (3072, 11, 22, 23);
      $Lhdr = 'a3av9a42a8v4Vv2V2v2Vv8Vv2Vv15a16v51V53a8v8V92v436Vv2V6v642';
      $Lrec = 'v4Vv2VvaaV2v14V30v8V2v274V180v544V2v48V2a78v9V5a52v12Vv102'; };
    defined($reclen) or die "$name\: Instrument $instrument not supported\nAbort";
    $Bhdr = $Lhdr; $Bhdr =~ s/v/n/g; $Bhdr =~ s/V/N/g;
    $Brec = $Lrec; $Brec =~ s/v/n/g; $Brec =~ s/V/N/g;
    open my $fh, "<$infile"; binmode $fh;
    $count = read($fh, $buf, $reclen);
    ($count == $reclen) or die "$name\: Header count != reclen $infile\nAbort";
    my @data = unpack($Lhdr, $buf);
    ($year, $doy) = ($data[$iyear], $data[$idoy]);
    my $Lendian = ($year > 1960 and $year < 2099 and $doy > 0 and $doy < 367);
    @data = unpack($Bhdr, $buf);
    ($year, $doy) = ($data[$iyear], $data[$idoy]);
    my $Bendian = ($year > 1960 and $year < 2099 and $doy > 0 and $doy < 367);
    ($Lendian xor $Bendian) or die "$name\: $infile endian indeterminate\nAbort";
    my $outfile = $data[$iout];
    $outfile =~ s/^\s+//; $outfile =~ s/\s+$//; $outfile =~ s/ //g;
    (length($outfile) != 42) and die "$name\: L1B filename $outfile funky\nAbort";
    open my $FH, ">$outfile"; binmode $FH;
    ($Lendian) and $buf = pack($Bhdr, unpack($Lhdr, $buf));
    print $FH $buf;
    while (!eof($fh)) {
      $count = read($fh, $buf, $reclen);
      ($count == $reclen) or die "$name\: Record count != reclen $infile\nAbort";
      ($Lendian) and $buf = pack($Brec, unpack($Lrec, $buf));
      print $FH $buf;
    }
    close $fh; close $FH;
    return $outfile;
  }

# ----------------------------
  sub populate_swath_info_jpss
# ============================
# @swath = populate_swath_info_jpss(\%dynanc_loc, %options);
  {
    # %dynanc_loc is remote/local information for GFS dynamic ancillary data
    # %options are the command line & default general processing options
    # @swath is an array of hash references, each reference points to granule
    # processing information, and this will finally be expressed as changes
    # to the PCF templates in $tmpdir. During the execution of this function,
    # the $PWD is assumed to be $tmpdir/swath_nnn.
    my %dynanc_loc = %{(shift)};
    my (%options) = @_;
    my $name = (caller(0))[3];
    $name =~ s/main/$MAIN/g;
    #
    # Patterns needed for extracting date/time from filenames
    # ```````````````````````````````````````````````````````
    my $sdr1 = 'XXXXX_xxx_d%Y%m%d_t%H%M%Sn_ennnnnnn';
    my $sdr2 = 'XXXXX_xxx_d%Y%m%d_tnnnnnnn_e%H%M%Sn';
    #
    # Variables needed for date/time tests
    # ````````````````````````````````````
    my ($YYYY, $MM, $DD, $hh, $mm, $ss) = undef;
    #
    # Push each valid (& thus fully populated) %granule in here
    # `````````````````````````````````````````````````````````
    my @swath = ();
    #
    # Check for microwave-included mode
    # `````````````````````````````````
    my (@atms_geo, @atms_beg, @atms_end, @atms_tdr) = ((), (), (), ());
    (grep {$_ eq $options{mode}} qw/HR_IRMW LR_IRMW/) and do {
      #
      # Make co-arrays of ATMS geo filenames and their data beg/end times
      # `````````````````````````````````````````````````````````````````
      @atms_tdr = (sort(glob('TATMS*')));
      foreach (sort(glob('GATMO*'))) {
        ($YYYY, $MM, $DD, $hh, $mm, $ss) = get_datetime_from_tokens($_, $sdr1);
        is_plausible_datetime($YYYY, $MM, $DD, $hh, $mm, $ss) or
          die "$name\: Date from ATMS GEO file $_ implausible\nAbort";
        my $beg = timegm_nocheck($ss, $mm, $hh, $DD, $MM-1, $YYYY);
        ($YYYY, $MM, $DD, $hh, $mm, $ss) = get_datetime_from_tokens($_, $sdr2);
        is_plausible_datetime($YYYY, $MM, $DD, $hh, $mm, $ss) or
          die "$name\: Date from ATMS GEO file $_ implausible\nAbort";
        my $end = timegm_nocheck($ss, $mm, $hh, $DD, $MM-1, $YYYY);
        # If $end precedes $beg it is because end time is in next day, so add a day
        ($end < $beg) and $end = timegm_nocheck($ss, $mm, $hh, $DD+1, $MM-1, $YYYY);
        push(@atms_geo, $_); push(@atms_beg, $beg); push(@atms_end, $end);
      }
    };
    #
    # If Clouds selected, I need to assemble filenames and timing so that I
    # can populate the HEAP_Clouds.PCF with CLOUD_GEO_FILE_x & CLOUD_FILE_y/z
    # ```````````````````````````````````````````````````````````````````````
    my (@viirs_mask, @viirs_height) = ((), ());
    my (@viirs_geo, @viirs_beg, @viirs_end) = ((), (), ());
    if (index($options{output}, 'C') != -1) {
      my $fmt = "%04d%02d%02d%02d%02d%02d";
      my $maskfmt = "JRR-CloudMask_*_%s_%s_%s_*.nc";
      my $heightfmt = "JRR-CloudHeight_*_%s_%s_%s_*.nc";
      foreach (sort(glob('GMTCO*'))) {
        ($YYYY, $MM, $DD, $hh, $mm, $ss) = get_datetime_from_tokens($_, $sdr1);
        is_plausible_datetime($YYYY, $MM, $DD, $hh, $mm, $ss) or
          die "$name\: Date from VIIRS GEO file $_ implausible\nAbort";
        my $beg = timegm_nocheck($ss, $mm, $hh, $DD, $MM-1, $YYYY);
        my $startmask = sprintf("s$fmt?", $YYYY, $MM, $DD, $hh, $mm, $ss);
        ($YYYY, $MM, $DD, $hh, $mm, $ss) = get_datetime_from_tokens($_, $sdr2);
        is_plausible_datetime($YYYY, $MM, $DD, $hh, $mm, $ss) or
          die "$name\: Date from VIIRS GEO file $_ implausible\nAbort";
        my $end = timegm_nocheck($ss, $mm, $hh, $DD, $MM-1, $YYYY);
        my $endmask = sprintf("e$fmt?", $YYYY, $MM, $DD, $hh, $mm, $ss);
        push(@viirs_geo, $_); push(@viirs_beg, $beg); push(@viirs_end, $end);
        my $maskglob = sprintf($maskfmt, $options{sat}, $startmask, $endmask);
        my $heightglob = sprintf($heightfmt, $options{sat}, $startmask, $endmask);
        my @maskfile = sort(glob($maskglob));
        my @heightfile = sort(glob($heightglob));
        if ($#maskfile != 0) {
          warn "$name\: WARNING: No JRR-CloudMask match to $_\n";
        } else { push(@viirs_mask, $maskfile[0]); }
        if ($#heightfile != 0) {
          warn "$name\: WARNING: No JRR-CloudHeight match to $_\n";
        } else { push(@viirs_height, $heightfile[0]); }
      }
    }
    # Loop over the candidate CrIS data in $PWD
    # `````````````````````````````````````````
    my @cris_sdr = (sort(glob('SCRI*')));
    foreach (sort(glob('GCRSO*'))) {
      # Check for companion SCRI? file for this GCRSO file
      my $tag = substr($_, 6, 31);
      my ($jcs, $jce) = job_coverage_start_end_jpss($_);
      my %granule = (JCS => $jcs, JCE => $jce, IR_GEO_FILE => $_);
      my $cris_sdr = 'N/A';
      foreach (@cris_sdr) {
        (substr($_, 6, 31) eq $tag) and $cris_sdr = $_;
      }
      ($cris_sdr eq 'N/A') and do {
        warn "$name\: WARNING: No SCRIS match to $_\n";
        next;
      };
      $granule{IR_FILE} = $cris_sdr;
      #
      # Get the beg/end data times for this CrIS granule
      # ````````````````````````````````````````````````
      ($YYYY, $MM, $DD, $hh, $mm, $ss) = get_datetime_from_tokens($_, $sdr1);
      is_plausible_datetime($YYYY, $MM, $DD, $hh, $mm, $ss) or
        die "$name\: Date from CRIS GEO file $_ implausible\nAbort";
      my $beg = timegm_nocheck($ss, $mm, $hh, $DD, $MM-1, $YYYY);
      ($YYYY, $MM, $DD, $hh, $mm, $ss) = get_datetime_from_tokens($_, $sdr2);
      is_plausible_datetime($YYYY, $MM, $DD, $hh, $mm, $ss) or
        die "$name\: Date from CRIS GEO file $_ implausible\nAbort";
      my $end = timegm_nocheck($ss, $mm, $hh, $DD, $MM-1, $YYYY);
      my $irmid = ($beg + $end)/2;
      #
      # This test tells me if I need to match up GATMO & TATMS files
      # ````````````````````````````````````````````````````````````
      my $jdx = 0; # initialize in case $jdx = $j never happens below
      (grep {$_ eq $options{mode}} qw/HR_IRMW LR_IRMW/) and do {
        # Use a mid point CrIS time to get the closest ATMS file
        my $mid = ($beg + $end) / 2;
        # Do we have trio of fore-, mid-, and aft- ATMS geo files?
        for (my $j=0; $j<=$#atms_geo; $j++) {
          (($mid > $atms_beg[$j]) and ($mid < $atms_end[$j])) and $jdx = $j;
        }
        (($jdx < 1) or ($jdx >= $#atms_geo)) and next;
        $granule{MW_GEO_FILE_1} = $atms_geo[$jdx-1];
        $granule{MW_GEO_FILE_2} = $atms_geo[$jdx];
        $granule{MW_GEO_FILE_3} = $atms_geo[$jdx+1];
        # Check for companion TATMS file for each of the above GATMO files
        my @tdr = ();
        for (my $j=$jdx-1; $j<=$jdx+1; $j++) {
          $tag = substr($atms_geo[$j], 6, 31);
          my $tdr = 'N/A';
          foreach (@atms_tdr) {
            (substr($_, 6, 31) eq $tag) and $tdr = $_;
          }
          ($tdr eq 'N/A') and do {
            ($options{verbose} >= 1) and
              warn "$name\: WARNING No TATMS match to $atms_geo[$j]\n";
            next;
          };
          push(@tdr, $tdr);
        }
        ($#tdr == 2) or next;
        $granule{MW_FILE_1} = $tdr[0];
        $granule{MW_FILE_2} = $tdr[1];
        $granule{MW_FILE_3} = $tdr[2];
      };
      #
      # This test tells me if I need to match up GMTCO & JRR files
      # ``````````````````````````````````````````````````````````
      $granule{USE_CLOUD_DATA_FLAG} = "NO";
      if (index($options{output}, 'C') != -1) {
        $granule{USE_CLOUD_DATA_FLAG} = "YES";
        my $k = 0; # $k is the index number used for CLOUD_GEO_FILE_$k
        for (my $j=0; $j<=$#viirs_geo; $j++) {
          #
          # In HEAP_Clouds.PCF I see CLOUD_PADDING=14_seconds. Take a punt
          # that is a time range around SCRIS/F in cloud overlap computation.
          # `````````````````````````````````````````````````````````````````
          ($viirs_end[$j] < ($beg-140)) and next;
          ($viirs_beg[$j] > ($end+140)) and next;
          $k++; my $m = 2*$k; my $l = $m-1;
          my $key1 = sprintf("CLOUD_GEO_FILE_%d", $k);
          my $key2 = sprintf("CLOUD_FILE_%d", $l);
          my $key3 = sprintf("CLOUD_FILE_%d", $m);
          $granule{"$key1"} = $viirs_geo[$j];
          $granule{"$key2"} = $viirs_height[$j];
          $granule{"$key3"} = $viirs_mask[$j];
        }
      }
      # GRANULE_INFO_FILE, IR_NC_FILE & NUCAPS_ALL_PRODUCT have creation times in their
      # filenames, so I cannot know the name until after I run HEAP_Preprocessor.pl.
      # I put in the PCFs a wildcard for each that I will later use to replace itself.
      # ``````````````````````````````````````````````````````````````````````````````
      my $globit = sprintf("s%s_e%s", $jcs, $jce);
      $granule{GRANULE_INFO_FILE} = sprintf("NUCAPS-GranuleInfo*%s*.txt", $globit);
      $granule{IR_NC_FILE} = sprintf("NUCAPS-XXX*%s*.nc", $globit);
      #
      # Before accepting this granule, check ATMS file contiguity
      # `````````````````````````````````````````````````````````
      my $contiguous = 1;
      (grep {$_ eq $options{mode}} qw/HR_IRMW LR_IRMW/) and do {
        $end = $atms_end[$jdx-1];
        for (my $j=$jdx; $j<=$jdx+1; $j++) {
          $beg = $atms_beg[$j];
          my $delta = $beg - $end;
          ($delta > 2) and $contiguous = 0;
          ($delta < 0) and $contiguous = 0;
          $end = $atms_end[$j];
        }
      };
      ($contiguous) or next;
      #
      # Load AVN_FILE1/2 with dummy values in case we are skipping the retrieval
      # `````````````````````````````````````````````````````````````````````````
      $granule{GFS_FILE_1} = 'DUMMY_GFS_FILE_1';
      $granule{GFS_FILE_2} = 'DUMMY_GFS_FILE_2';
      if (index($options{output}, 'R') != -1) {
        my ($avn1, $avn2) = gfs_fetch_with_fallback($irmid, \%dynanc_loc, %options);
        -e $avn1 or next; # no AVN_FILE1 so we bail on this granule
        -e $avn2 or next; # no AVN_FILE2 so we bail on this granule
        #
        # All checks complete, push it onto swath processing stack
        # ````````````````````````````````````````````````````````
        $granule{GFS_FILE_1} = $avn1;
        $granule{GFS_FILE_2} = $avn2;
      }
      ($options{verbose} == 3) and pretty_dump('%granule', %granule);
      push (@swath, \%granule);
    }
    return @swath;
  }

# -----------------------------
  sub populate_swath_info_metop
# =============================
# @swath = populate_swath_info_metop(\%dynanc_loc, %options);
  {
    # %dynanc_loc is remote/local information for GFS dynamic ancillary data
    # %options are the command line & default general processing options
    # @swath is an array of hash references, each reference points to granule
    # processing information, and this will finally be expressed as changes
    # to the PCF templates in $tmpdir. During the execution of this function,
    # the $PWD is assumed to be $tmpdir/swath_nnn.
    my %dynanc_loc = %{(shift)};
    my (%options) = @_;
    my $name = (caller(0))[3];
    $name =~ s/main/$MAIN/g;
    #
    # Patterns needed for extracting date/time from filenames
    # ```````````````````````````````````````````````````````
    my $nss1 = 'NSS.xxxX.Mx.D%y%j.S%H%M.E1444';
    my $nss2 = 'NSS.xxxX.Mx.D%y%j.Sxxxx.E%H%M';
    my $iasi1 = 'IASI_xxx_1C_M0x_%Y%m%d%H%M%SZ_YYYYMMDDHHMMSSZ'; 
    my $iasi2 = 'IASI_xxx_1C_M0x_YYYYMMDDHHMMSSZ_%Y%m%d%H%M%SZ'; 
    #
    # Variables needed for date/time tests
    # ````````````````````````````````````
    my ($YYYY, $MM, $DD, $DOY, $hh, $mm, $ss, $tod) = undef;
    #
    # Loop over the IASI L1C files in $PWD
    # ````````````````````````````````````
    my @swath = ();
    foreach my $iasi_l1c (sort(glob('IASI_xxx_1C_M*'))) {
      ($YYYY, $MM, $DD, $hh, $mm, $ss) = get_datetime_from_tokens($iasi_l1c, $iasi1);
      is_plausible_datetime($YYYY, $MM, $DD, $hh, $mm, $ss) or
        die "$name\: Date from MetOp IASI L1C file $_ implausible\nAbort";
      my $beg = timegm_nocheck($ss, $mm, $hh, $DD, $MM-1, $YYYY);
      my $jcs = $YYYY.$MM.$DD.$hh.$mm.$ss.'0';
      ($YYYY, $MM, $DD, $hh, $mm, $ss) = get_datetime_from_tokens($iasi_l1c, $iasi2);
      is_plausible_datetime($YYYY, $MM, $DD, $hh, $mm, $ss) or
        die "$name\: Date from MetOp IASI L1C file $_ implausible\nAbort";
      my $end = timegm_nocheck($ss, $mm, $hh, $DD, $MM-1, $YYYY);
      my $irmid = ($beg + $end)/2;
      my $jce = $YYYY.$MM.$DD.$hh.$mm.$ss.'0';
      my %granule = (JCS => $jcs, JCE => $jce, IR_FILE => $iasi_l1c);
      $granule{USE_CLOUD_DATA_FLAG} = "NO";
      #
      # So this test tells me if I need MW files (i.e. AMSUA/MHS)
      # `````````````````````````````````````````````````````````
      my ($j, $k) = (0, 1);
      (grep {$_ eq $options{mode}} qw/S_IRMW M_IRMW E_IRMW/) and do {
        foreach (sort(glob('NSS.*'))) {
          (substr($_, 0, 4) ne 'NSS.') and next;
          $YYYY = '20'.token_value('%y', $_, $nss1);
          $DOY = token_value('%j', $_, $nss1);
          $hh = token_value('%H', $_, $nss1);
          $mm = token_value('%M', $_, $nss1);
          is_plausible_doytime($YYYY, $DOY, $hh, $mm) or
            die "$name\: DoyTime from MetOp NSS. file $_ implausible\nAbort";
          $tod = $hh.$mm; # time of day as hhmm
          my $begmw = timegm_nocheck(0, $mm, $hh, $DOY, 0, $YYYY);
          $hh = token_value('%H', $_, $nss2);
          $mm = token_value('%M', $_, $nss2);
          is_plausible_doytime($YYYY, $DOY, $hh, $mm) or
            die "$name\: DoyTime from MetOp NSS. file $_ implausible\nAbort";
          ($tod gt $hh.$mm) and $DOY++;
          my $endmw = timegm_nocheck(0, $mm, $hh, $DOY, 0, $YYYY);
          ($endmw < $beg) and next;
          ($begmw > $end) and next;
          (index($_, '.AMAX.') != -1) and do {
            $j++; my $mw = sprintf("MW_FILE_%d", $j++); $granule{"$mw"} = $_; };
          (index($_, '.MHSX.') != -1) and do {
            $k++; my $mw = sprintf("MW_FILE_%d", $k++); $granule{"$mw"} = $_; };
        }
      };
      # GRANULE_INFO_FILE, IR_NC_FILE & NUCAPS_ALL_PRODUCT have creation times in their
      # filenames, so I cannot know the name until after I run HEAP_Preprocessor.pl.
      # I put in the PCFs a wildcard for each that I will later use to replace itself.
      # ``````````````````````````````````````````````````````````````````````````````
      my $globit = sprintf("s%s_e%s", $jcs, $jce);
      $granule{GRANULE_INFO_FILE} = sprintf("NUCAPS-GranuleInfo*%s*.txt", $globit);
      $granule{IR_NC_FILE} = sprintf("NUCAPS-XXX*%s*.nc", $globit);
      $granule{AMSUA_NC_FILE} = sprintf("AMSUA-ALL*%s*.nc", $globit);
      $granule{MHS_NC_FILE} = sprintf("MHS-ALL*%s*.nc", $globit);
      #
      # Load AVN_FILE1/2 with dummy values in case we are skipping the retrieval
      # ````````````````````````````````````````````````````````````````````````
      $granule{GFS_FILE_1} = 'DUMMY_GFS_FILE_1';
      $granule{GFS_FILE_2} = 'DUMMY_GFS_FILE_2';
      if (index($options{output}, 'R') != -1) {
        my ($avn1, $avn2) = gfs_fetch_with_fallback($irmid, \%dynanc_loc, %options);
        -e $avn1 or next; # no AVN_FILE1 so we bail on this granule
        -e $avn2 or next; # no AVN_FILE2 so we bail on this granule
        #
        # All checks complete, push it onto swath processing stack
        # ````````````````````````````````````````````````````````
        $granule{GFS_FILE_1} = $avn1;
        $granule{GFS_FILE_2} = $avn2;
      }
      ($options{verbose} == 3) and pretty_dump('%granule', %granule);
      push (@swath, \%granule);
    }
    return @swath;
  }

# ---------------
  sub pretty_dump
# ===============
# pretty_dump($title, %hash);
  {
    # Do a "pretty print" of %hash by effectively grepping Dumper().
    # Assumes standard module Data::Dumper has been loaded.
    my $title = 'DEBUG '.shift;
    my (%hash) = @_;
    print '-' x length($title)."\n";
    print "$title\n";
    print '=' x length($title)."\n";
    foreach (split('\n', Dumper(\%hash))) {
      chomp;
      ((/VAR1/) or (/\}\;/)) and next;
      $_ =~ s/  //g; # shorten up the whitespace
      print "$_\n";
    }
    print "$title _END_\n";
  }

# ---------------
  sub token_value
# ===============
# $value = token_value($token, $file, $mask);  ...example follows...
# $value = token_value('%j', 't1.11332.1500.mod04.hdf', 't1.%y%j.%H%M.mod04.hdf');
  {
    # Takes the filename $file (within which tokens have already been expanded),
    # and the $mask (within which tokens are unexpanded) and, for a given $token
    # return the $value extracted from $file. Here is an example ...
    #   $value = token_value('%j', 't1.11332.1500.mod04.hdf', 't1.%y%j.%H%M.mod04.hdf');
    # which would return $value as '332'.
    # These are the supported tokens, returns the first occurance if more than one:
    #   %Y  Year with century. (1582-9999)
    #   %y  Year without century (00–99).
    #   %m  Month as decimal number (01–12).
    #   %j  Day of year as decimal number (001–366).
    #   %d  Day of the month as decimal number (01–31).
    #   %H  Hours as decimal number (00–23).
    #   %M  Minute as decimal number (00–59).
    #   %S  Second as decimal number (00–59).
    #   %F  3-digit forecast hour (000-999).
    #   %f  2-digit forecast hour (00-99).
    my $token = shift;
    my $file = shift;
    my $mask = shift;
    my $value = undef;
    my $expand = $mask;
    my @tokens = qw/  %Y %y %m  %j %d %H %S %M  %F %f/;
    my @widths =   (   4, 2, 2,  3, 2, 2, 2, 2,  3, 2);
    my $width = 0;
    for (my $j=0; $j<=$#tokens; $j++) {
      ($token eq $tokens[$j]) and do { $width = $widths[$j]; next; };
      my $blank = 'X' x $widths[$j];
      $expand =~ s/$tokens[$j]/$blank/g; 
    }
    my $i = index($expand, $token);
    if ($i > -1) {
      $value = substr($file, $i, $width);
    }
    return $value;
  }

# ---------------
  sub trim_string
# ===============
# $new_string = trim_string($old_string);
  {
    my $string = shift;
    $string =~ s/^\s+//;
    $string =~ s/\s+$//;
    return $string;
  }

# ---------
  sub usage
# =========
# usage() or usage($message)
  {
    # The help screen is now so long that it is confusing to put an error
    # or warning message to stdout before the help information because it
    # scrolls out of view.  So an optional messsage is now printed at end.
    my $msg = ''; ($#_ != -1) and $msg = $_[0];
    my $name = (caller(0))[3];
    $name =~ s/main/$MAIN/g;
    open(my $fh, "<$0");
    while (<$fh>) { chomp;
      if ( /START-HELP/ ) {
        while (<$fh>) { chomp;
          ( /END-HELP/ ) and do { close($fh); print $msg; exit(9); };
          $_ =~ s/\$CSPP_NUCAPS_HOME/$ENV{CSPP_NUCAPS_HOME}/g;
          my $help = (split(/\#/, $_))[1]; # remove leading hash
          $help and print "$help\n" or print "\n";
        }
      }
    }
    die "$name\: ERROR! Have you tinkered with help in script header?\nAbort";
  }

# --------------------
  sub validate_options
# ====================
# %options = validate_options($filemask, %options);
  {
    # $filemask is for globbing the nominated satellite data.
    # %options are the command line options to "validate".
    # Fatal errors in here terminate with die without showing usage() message.
    # The most important task is to convert $options{input_files} to a scalar
    # $options{input_dir} and filename list put back in $options{input_files}.
    # This is also where the working and ancillary data directories are made
    # if they do not exist, and their writable status verified.
    my $filemask = shift;
    my (%options) = @_;
    my $name = (caller(0))[3];
    $name =~ s/main/$MAIN/g;
    my $globber = $options{input_files};
    (-d $options{input_files}) and do {
      my $base = $options{input_files};
      $globber = File::Spec -> catfile($base, $filemask); };
    my @input_files = sort(glob($globber));
    (($#input_files == 0) and !(-e $input_files[0])) and
      die "$name\: INPUT_FILES $input_files[0] not found. Abort\n";
    ($#input_files < 0) and die "$name\: INPUT_FILES not found. Abort\n";
    $options{input_dir} = dirname(abs_path($input_files[0]));
    $options{input_files} = ();
    foreach (@input_files) {
      my $filename = basename($_);
      (index($filename, '.xml') != -1) and next; # pesky xmls may be intermingled
      push(@{$options{input_files}}, $filename);
    }
    ($options{gfs} > 3) and do {
      die "$name\: --gfs=GFS $options{gfs} is not valid.\nAbort"; };
    ($options{sat} eq 'npp') and do {
      ($options{fovs} < 7 or $options{fovs} > 9) and do {
        die "$name\: --fovs=N_FOVS $options{fovs} is not valid.\nAbort"; }; };
    (-d $options{work}) or do {
      ($options{verbose} >=2) and print "Making WORK dir $options{work}\n";
      mkpath($options{work}) or
        die "$name\: WORK dir $options{work} cannot be made.\nAbort"; };
    (-w $options{work}) or
      die "$name\: WORK dir $options{work} not writable.\nAbort";
    #
    # If Retrieval NOT selected, we do not need to fetch GFS ancillary data
    # `````````````````````````````````````````````````````````````````````
    if (index($options{output}, 'R') == -1) { $options{quarantine} = 1; }
    ($options{quarantine}) or do {
      (-d $options{dynanc}) or do {
        ($options{verbose} >=2) and print "Making dynanc DIR $options{dynanc}\n";
        mkpath($options{dynanc}) or
          die "$name\: DYNANC dir $options{dynanc} cannot be made.\nAbort"; };
      (-w $options{dynanc}) or
        die "$name\: DYNANC dir $options{dynanc} not writable.\nAbort"; };
    return %options;
  }
